# **********************************************************************
# Copyright 2018 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import nsx_logging
try:
    import simplejson
except ImportError:
    import json as simplejson

import cli.commands.host_shared.ovs_util as ovs_util
from cli.commands.cli_command_provider import CliCommandProvider

lg = nsx_logging.getLogger("cli.commands.host_kvm.ip_discovery")


class GetIpDiscoveryConfig(CliCommandProvider):
    def __call__(self, cmd_list):
        configs = []
        lport = None
        num_args = len(cmd_list)
        if (num_args > 3):
            lport = cmd_list[3]
        output = ovs_util.get_ip_discovery_config(lport)
        if (output is None):
            return configs
        for conf in output:
            config = {}
            config['logical_port'] = conf[0]
            config['arp_snooping'] = conf[1]
            config['dhcp_snooping'] = conf[2]
            config['nd_snooping'] = conf[3]
            config['dhcpv6_snooping'] = conf[4]
            config['arp_nd_expiry_timeout'] = conf[7] + " min"
            configs.append(config)
        return "OK\n\n%s" % simplejson.dumps(configs)


class GetIpDiscoveryEntry(CliCommandProvider):
    def __call__(self, cmd_list):
        lport = None
        version = None
        num_args = len(cmd_list)
        if (num_args > 3):
            lport = cmd_list[3]
            if (num_args > 4):
                version = cmd_list[4]
        active_entries = self._get_active_entries(lport, version)
        inactive_entries = self._get_inactive_entries(lport, version)
        return "OK\n\n%s" % simplejson.dumps(active_entries + inactive_entries)

    def _get_active_entries(self, lport, version):
        active_entries = []
        output = ovs_util.get_ip_discovery_active_entries(lport, version)
        if output is None:
            return active_entries
        for entry in output:
            active_entry = {}
            active_entry['discovery_type'] = entry[1]
            active_entry['logical_port'] = entry[2]
            active_entry['ip_address'] = entry[3]
            active_entry['mac_address'] = entry[4]
            active_entry['vlan'] = entry[5]
            active_entry['timeout'] = entry[6]
            active_entries.append(active_entry)
        return active_entries

    def _get_inactive_entries(self, lport, version):
        inactive_entries = []
        output = ovs_util.get_ip_discovery_inactive_entries(lport, version)
        if output is None:
            return inactive_entries
        for entry in output:
            inactive_entry = {}
            inactive_entry['discovery_type'] = entry[1]
            inactive_entry['logical_port'] = entry[2]
            inactive_entry['ip_address'] = entry[3]
            inactive_entry['mac_address'] = entry[4]
            inactive_entry['vlan'] = entry[5]
            inactive_entry['timeout'] = "Expired"
            inactive_entries.append(inactive_entry)
        return inactive_entries


class GetIpDiscoveryStats(CliCommandProvider):
    def __call__(self, cmd_list):
        ip_discovery_stats = []
        lport = None
        num_args = len(cmd_list)
        if (num_args > 3):
            lport = cmd_list[3]
        output = ovs_util.get_ip_discovery_stats(lport)
        dhcp_pending_reqs_stats = ovs_util.get_dhcp_pending_reqs_stats(lport)
        if output is None or dhcp_pending_reqs_stats is None:
            return ip_discovery_stats
        for entry in output:
            ip_discovery_stats_entry = {}
            ip_discovery_stats_entry['logical_port'] = entry[0]
            ip_discovery_stats_entry['max_arp_entries'] = entry[1]
            ip_discovery_stats_entry['active_arp_entries'] = entry[2]
            ip_discovery_stats_entry['expired_arp_entries'] = entry[3]
            ip_discovery_stats_entry['max_dhcpv4_entries'] = entry[4]
            ip_discovery_stats_entry['total_dhcpv4_entries'] = entry[5]
            ip_discovery_stats_entry['max_nd_entries'] = entry[6]
            ip_discovery_stats_entry['active_nd_entries'] = entry[7]
            ip_discovery_stats_entry['expired_nd_entries'] = entry[8]
            ip_discovery_stats_entry['max_dhcpv6_entries'] = entry[9]
            ip_discovery_stats_entry['total_dhcpv6_entries'] = entry[10]
            pending_request_entry = [p for p in dhcp_pending_reqs_stats
                                     if p[0] == entry[0]][0]
            ip_discovery_stats_entry['max_pending_dhcpv4_reqs'] = \
                pending_request_entry[1]
            ip_discovery_stats_entry['pending_dhcpv4_reqs'] = \
                pending_request_entry[2]
            ip_discovery_stats_entry['max_pending_dhcpv6_reqs'] = \
                pending_request_entry[3]
            ip_discovery_stats_entry['pending_dhcpv6_reqs'] = \
                pending_request_entry[4]
            ip_discovery_stats.append(ip_discovery_stats_entry)
        return "OK\n\n%s" % simplejson.dumps(ip_discovery_stats)


class GetIpDiscoveryIgnoreList(CliCommandProvider):
    def __call__(self, cmd_list):
        ignore_list = []
        lport = None
        version = None
        num_args = len(cmd_list)
        if (num_args > 3):
            lport = cmd_list[3]
            if (num_args > 4):
                version = cmd_list[4]
        output = ovs_util.get_ip_discovery_ignore_list(lport, version)
        if output is None:
            return ignore_list
        for entry in output:
            ignore_list_entry = {}
            ignore_list_entry['logical_port'] = entry[1]
            ignore_list_entry['ip_address'] = entry[2]
            ignore_list_entry['mac_address'] = entry[3]
            ignore_list_entry['vlan'] = entry[4]
            ignore_list.append(ignore_list_entry)
        return "OK\n\n%s" % simplejson.dumps(ignore_list)


class GetIpDiscoveryIgnoreListStats(CliCommandProvider):
    def __call__(self, cmd_list):
        ignore_list_stats = []
        lport = None
        num_args = len(cmd_list)
        if (num_args > 4):
            lport = cmd_list[4]
        output = ovs_util.get_ip_discovery_ignore_list_stats(lport)
        if output is None:
            return ignore_list_stats
        for entry in output:
            ignore_list_stats_entry = {}
            ignore_list_stats_entry['logical_port'] = entry[0]
            ignore_list_stats_entry['v4_ignore_count'] = entry[1]
            ignore_list_stats_entry['v6_ignore_count'] = entry[2]
            ignore_list_stats.append(ignore_list_stats_entry)
        return "OK\n\n%s" % simplejson.dumps(ignore_list_stats)
