# **********************************************************************
# Copyright 2018 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import json
import nsx_logging
import re

from cli.commands.cli_command_provider import CliCommandProvider
from cli.utils.command_utils import invoke_command
from cli.utils.error_code import ErrorCode as ec
from cli.utils.location_utils import FilePaths

lg = nsx_logging.getLogger("cli.commands.kvm.lldp")

OVS_APPCTL = '/usr/bin/ovs-appctl'
LLDP_SOCKET = '/var/run/vmware/nsx-opsagent/opsagent-cli'

ENABLED = 'enabled'
DISABLED = 'disabled'

INTERFACE = 'Interface'

KEY_MAPPING = {'Interface': 'Chassis Id',
               'Mac': 'Port Id',
               'Port description': 'Port Description',
               'System name': 'System Name',
               'System description': 'System Description',
               'Enabled capabilities': 'Enabled Capabilities',
               'Management address': 'Management Address',
               'System port number': 'System Port Number',
               'Unknown interface numbering': 'Unknown',
               'Index': 'ifIndex',
               'Oid string': 'OID',
               'Mac/phy auto-negotiation support': 'Mac/phy Auto-negotiation Support',
               'Mac/phy auto-negotiation status': 'Mac/phy Auto-negotiation Status',
               'Mac/phy auto-negotiation capability':'Mac/phy Auto-negotiation Capability',
               'Mac/phy MAU type': 'Mac/phy MAU Type',
               'MDI power port class': 'MDI Power Port Class',
               'PSE MDI power support': 'PSE MDI Power Support',
               'PSE MDI power state': 'PSE MDI Power State',
               'PSE pairs control ability': 'PSE Pairs Control Ability',
               'PSE power pair': 'PSE Power Pair',
               'Power class': 'Power Class',
               'Aggregation capability': 'Aggregation Capability',
               'Aggregation status': 'Aggregation Status',
               'Aggregated port id': 'Aggregated Port ID',
               'Maximum frame size': 'Maximum Frame Size'
               }


def _get_field_value(lldp_string, field_key):
    startIdx = lldp_string.find(field_key)
    if startIdx == -1:
        return ''
    startIdx = lldp_string.find(field_key) + len(field_key + ': ')
    endIdx = lldp_string.find('\n', startIdx)
    if endIdx == -1:
        return lldp_string[startIdx:]
    else:
        return lldp_string[startIdx:endIdx]


def _get_lldp_interfaces():
    '''
    Get the valid interfaces on the host.
    '''
    idx = 0
    devices = []
    cmd = [OVS_APPCTL, "-t", LLDP_SOCKET, "lldpd/status"]
    rc, out, err = invoke_command(cmd)
    if rc != 0:
        return "MSG_CLI_UNEXPECTED_ERROR\n%s" % err

    startIdxs = []
    startIdx = 0
    cnt = out.count(INTERFACE)
    for i in range(cnt):
        if i == 0:
            idx = out.find('Name', startIdx)
        else:
            idx = out.find('Name', startIdx + len('Name'))
        startIdxs.append(idx)
        startIdx = idx
    startIdxs.append(-1)

    for i in range(cnt):
        startIdx = startIdxs[i]
        endIdx = startIdxs[i + 1]
        interface_str = out[startIdx:endIdx]
        devices.append(_get_field_value(interface_str, 'Name'))
    return devices


class KvmViewLldpInterfaceArgument(object):
    @staticmethod
    def verify(id_str, reg_dict):
        ifs = _get_lldp_interfaces()
        if id_str in ifs:
            return True
        return False


def _get_lldp_config(dev_name=None):
    '''
    Parse the output of the following command to get the config
    :~# ovs-appctl -t /var/run/vmware/nsx-opsagent/opsagent-cli lldpd/status
    Interface:
        Name: eth1
        Tx: disabled
        Rx: enabled
        Tx Interval: 30
        Neighbors count: 2
    Interface:
        Name: eth2
        Tx: enabled
        Rx: enabled
        Tx Interval: 30
        Neighbors count: 2
    '''
    cmd = [OVS_APPCTL, "-t", LLDP_SOCKET, "lldpd/status"]
    rc, out, err = invoke_command(cmd)
    if rc != 0:
        return "MSG_CLI_UNEXPECTED_ERROR\n%s" % err
    configSummary = []

    '''
    enabled for Rx, and disabled for Tx by default, and Rx is always
    enabled on KVM.
    '''
    startIdxs = []
    startIdx = 0
    cnt = out.count(INTERFACE)
    for i in range(cnt):
        if i == 0:
            idx = out.find(INTERFACE, startIdx)
        else:
            idx = out.find(INTERFACE, startIdx + len(INTERFACE))
        startIdxs.append(idx)
        startIdx = idx
    startIdxs.append(-1)

    for i in range(cnt):
        device = {}
        startIdx = startIdxs[i]
        endIdx = startIdxs[i + 1]
        interface_str = out[startIdx:endIdx]

        device['device'] = _get_field_value(interface_str, 'Name')
        device['tx'] = _get_field_value(interface_str, 'Tx')
        device['rx'] = _get_field_value(interface_str, 'Rx')
        device['tx_interval'] = _get_field_value(interface_str, 'Tx Interval')
        if dev_name is None or dev_name == device['device']:
            configSummary.append(device)

    return "OK\n\n%s" % json.dumps(configSummary)


class ShowLldpConfig(CliCommandProvider):

    def __call__(self, cmd_list):
        dev_name = self.get_arg_value("lldp-interface-name-kvm", cmd_list)
        return _get_lldp_config(dev_name)


def _add_into_neighbor(neighbor, k, v):
    if (v == ''):
        return
    neighbor[k] = v


def _get_neighbor_number(lldp_string):
    '''
    Both device and neighbors have Interface field,
    so we should minus one to get the neighbors count.
    '''
    return lldp_string.count(INTERFACE)


def _get_neighbors(lldp_string, cnt):
    neighbors = []
    startIdxs = []
    startIdx = 0
    for i in range(cnt):
        if i == 0:
            idx = lldp_string.find(INTERFACE, startIdx)
        else:
            idx = lldp_string.find(INTERFACE, startIdx + len(INTERFACE))
        startIdxs.append(idx)
        startIdx = idx
    startIdxs.append(-1)

    for i in range(cnt):
        neighbor = {}
        startIdx = startIdxs[i]
        endIdx = startIdxs[i+1]
        neighbor_str = lldp_string[startIdx:endIdx]

        for key in KEY_MAPPING:
            v = _get_field_value(neighbor_str, key)
            _add_into_neighbor(neighbor, KEY_MAPPING[key], v)

        neighbors.append(neighbor)

    return neighbors


def _get_device_info(cmd_list, dev_name):
    rc, out, err = invoke_command(cmd_list)
    if rc != 0:
        return "MSG_CLI_UNEXPECTED_ERROR\n%s" % err
    result = {}
    result['Device'] = dev_name
    idx = out.find('Interface') + len('Interface:')
    neighbor_cnt = _get_neighbor_number(out[idx:])
    result['Neighbor Count'] = neighbor_cnt
    result['Neighbors'] = _get_neighbors(out[idx:], neighbor_cnt)
    return result


def _get_lldp_neigh(dev_name=None):
    '''
    Parse the output of the following command to get the neighbors.
    # ovs-appctl -t /var/run/vmware/nsx-opsagent/opsagent-cli lldpd/show
      Interface:
      Name: eth1
      Mac: 00:50:56:98:a5:b4
      Index: 3
      Neighbors:
      Interface: eth1
      Mac: 00:50:56:98:d0:30
      Port description: driver_name: e1000e, driver_version: 3.2.6-k
                        firmware_version: 1.8-0
      System name: KVM-TN-01
      System description: x86_64 #1 SMP Sun Jan 14 10:36:03 EST 2018
      Capabilities: Bridge
      Enabled capabilities: Bridge
      Management address: ::80fe:0:217:5380:300:0
      Index: 3
      Maximum frame size: 1618
    '''
    result = []
    devices = []
    cmd_list = [OVS_APPCTL, "-t", LLDP_SOCKET, "lldpd/show"]
    if dev_name is not None:
        devices.append(dev_name)
    else:
        devices = _get_lldp_interfaces()

    for dev in devices:
        cmd_list.append(dev)
        result.append(_get_device_info(cmd_list, dev))
        cmd_list = cmd_list[0:-1]

    return "OK\n\n%s" % json.dumps(result)


class ShowLldpNeigh(CliCommandProvider):

    def __call__(self, cmd_list):
        dev_name = self.get_arg_value("lldp-interface-name-kvm", cmd_list)
        return _get_lldp_neigh(dev_name)
