# **********************************************************************
# Copyright 2017 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import nsx_logging
import json
import os
import sys

from cli.commands.cli_command_provider import CliCommandProvider
import cli.commands.host_shared.ovs_util as ovs_util
from cli.utils.command_utils import invoke_command
from cli.utils.location_utils import FilePaths
from cli.utils.error_code import ErrorCode as ec
import cli.utils.node_utils as node_utils
import cli.utils.file_utils as file_utils
from cli.utils.render_utils import get_text_pager

lg = nsx_logging.getLogger("cli.commands.hostpubliccloud.host_public_cloud")
OVERLAY = "overlay"
NSX_NAMESPACE = "nsx"
_SUDO = "/usr/bin/sudo"
_TMP_BUNDLE_DIR = "/tmp"

class DumpGWConnection(CliCommandProvider):
    def __call__(self, cmd_list):
        r = ovs_util.get_gateway_connection_status()
        if not r:
            return "MSG_PUBLIC_CLOUD_GW_FAILED"
        res = {}
        res['gw_addr'] = r[0].split(":", 1)[-1].strip()
        res['status'] = r[1].split(":", 2)[-1].strip()
        res['connect_time'] = r[2].split("#", 2)[-1].strip()
        res['last_known_error'] = r[3].split(":", 2)[-1].strip()
        return "OK\n\n%s" % json.dumps(res)


class DumpHostNetworkModeInterface(CliCommandProvider):
    def __call__(self, cmd_list):
        res = node_utils.get_public_cloud_config()
        if not res:
            return "MSG_PUBLIC_CLOUD_MODE_FAILED"
        lport_list = ovs_util.get_lport_list()
        if not lport_list:
            return "MSG_PUBLIC_CLOUD_MODE_FAILED"
        # Only 1 logical port for public cloud VM
        if "hybrid" in (val.lower() for val in lport_list[0]):
            res['mode'] = "Hybrid"
        return "OK\n\n%s" % json.dumps(res)


class DumpGWCertificates(CliCommandProvider):
    def __call__(self, cmd_list):
        if len(cmd_list) == 3:
            r = ovs_util.get_gateway_certificates("all")
        else:
            r = ovs_util.get_gateway_certificates(cmd_list[-1])
        if r is None:
            return "MSG_PUBLIC_CLOUD_GW_FAILED"
        output = []
        i = 0
        while i < len(r):
            res = {}
            res['ip'] = r[i].split(":", 1)[-1].strip()
            res['thumbprint'] = r[i+1].split(":", 1)[-1].strip()
            res['sha'] = r[i+2].split(":", 1)[-1].strip()
            i = i + 3
            output.append(res)
        return "OK\n\n%s" % json.dumps(output)


class DelGWCertificates(CliCommandProvider):
    def __call__(self, cmd_list):
        if len(cmd_list) == 3:
            r = ovs_util.del_gateway_certificates("all")
        else:
            r = ovs_util.del_gateway_certificates(cmd_list[-1])

        return ("OK\n\n%s" % json.dumps(""))


class ClearNsxVmAgentState(CliCommandProvider):
    def __call__(self, cmd_list):
        if node_utils.is_windows():
            cmd = ["powershell", "-ExecutionPolicy", "Bypass",
                   "-File", FilePaths.NSX_PUBLIC_CLOUD_BOOTSTRAPPER,
                   "-operation", "uninstall"]
        else:
            cmd = ["service", "nsx-agent", "stop"]
            rc, out, err = invoke_command(cmd)
            if rc:
                lg.error("Failed to stop nsx-agent: %s", err,
                         error_code=ec.OPERATION_FAILED)
                return "MSG_PUBLIC_CLOUD_CLEAR_STATE_FAILED"
            cmd = ["service", "nsx-agent", "clear-state"]

        rc, out, err = invoke_command(cmd)
        if rc:
            lg.error("Failed to clear nsx-agent state: %s", err,
                     error_code=ec.OPERATION_FAILED)
            return "MSG_PUBLIC_CLOUD_CLEAR_STATE_FAILED"
        return ("OK\n\n%s" % json.dumps(""))


class StartVmAgentUpgrade(CliCommandProvider):
    def __call__(self, cmd_list):
        new_version = cmd_list[2]
        old_version = node_utils.get_node_version()
        if (new_version == old_version):
            return "MSG_PUBLIC_CLOUD_UPGRADE_VERSION_SAME"

        download_from_pcg = False
        if (len(cmd_list) > 4):
            file_store = FilePaths.NSX_FILE_STORE
            bundle = cmd_list[4]
            path = os.path.join(file_store, bundle)
            if not os.path.isfile(path):
                lg.error("File %s is not present in %s",
                         bundle, file_store,
                         error_code=ec.FILE_NOT_FOUND)
                return "MSG_PUBLIC_CLOUD_UPGRADE_BUNDLE_NOT_FOUND"
        else:
            download_from_pcg = True

        if node_utils.is_windows():
            cmd = ["powershell", "-ExecutionPolicy", "Bypass",
                   "-File", FilePaths.NSX_UPGRADE_SCRIPT, "-upgrade",
                   "true"]
            if not download_from_pcg:
                cmd.extend(["-bundlePath", path])
            else:
                cmd.extend(["-downloadScript", "true"])
        else:
            cmd = [_SUDO]
            res = node_utils.get_public_cloud_config()
            if not res:
                return "MSG_PUBLIC_CLOUD_UPGRADE_FAILED"
            if res.get("mode") == OVERLAY:
                cmd.extend(["ip", "netns", "exec", NSX_NAMESPACE])
            cmd.extend([FilePaths.NSX_UPGRADE_SCRIPT, "--upgrade"])
            if not download_from_pcg:
                cmd.extend(["--bundle", path])

        rc, out, err = invoke_command(cmd)
        if (rc != 0):
            lg.error("Failed to start the upgrade, rc: %s, out: %s, err: %s",
                     rc, out, err, error_code=ec.OPERATION_FAILED)
            return "MSG_PUBLIC_CLOUD_UPGRADE_FAILED"

        res = {}
        res['status'] = "Started"
        return ("OK\n\n%s" % json.dumps(res))

class DumpGwFullSyncState(CliCommandProvider):
    def __call__(self, cmd_list):
        r = ovs_util.get_gateway_fullsync_status()
        if not r:
            return "MSG_PUBLIC_CLOUD_FULL_SYNC_FAILED"
        res = {}
        res['status'] = r[0].split(":", 1)[-1].strip()
        return "OK\n\n%s" % json.dumps(res)

class GetSupportBundleFiles(CliCommandProvider):
    def __call__(self, cmd_list):
        # This version of the command saves the bundle to the filestore
        display = False
        if node_utils.is_windows():
            cmd = ["powershell", "-ExecutionPolicy", "Bypass",
                   "-File", FilePaths.SUPPORT_BUNDLE_SCRIPT,
                   "-nsxBundleOutputDir", FilePaths.NSX_FILE_STORE]
            if len(cmd_list) is 4:
                if os.path.exists(os.path.join(
                                  FilePaths.NSX_FILE_STORE, cmd_list[3])):
                    return "MSG_FILESTORE_ALREADY_EXISTS\n%s" % cmd_list[3]
                cmd.extend(["-bundleName", cmd_list[3]])
        else:
            cmd = [_SUDO, "ip", "netns", "exec", NSX_NAMESPACE,
                   FilePaths.SUPPORT_BUNDLE_SCRIPT]
            if len(cmd_list) is 4:
                dst_filename = cmd_list[3]
                if dst_filename is not None and os.path.exists(
                    os.path.join(file_utils.get_file_store(), dst_filename)):
                    return "MSG_FILESTORE_ALREADY_EXISTS\n%s" % dst_filename
            elif len(cmd_list) is 2:
                cmd.append("-o True")
                display = True

        err_details = "internal server error"
        try:
            rc, out, err = invoke_command(cmd, display=display)
            if rc == 0:
                if node_utils.is_windows() is False and len(cmd_list) is 4:
                    generated_bundle_filename = out.strip().split("/")[-1]
                    # Move the generated support-bundle to the filestore
                    result, err_msg_id = file_utils.move_file_to_filestore(
                        generated_bundle_filename, dst_filename,
                        src_dir=_TMP_BUNDLE_DIR)
                    if result:
                        return "OK\n\n%s" % json.dumps(
                            {"filename": dst_filename,
                            "path": FilePaths.NSX_FILE_STORE})
                    elif err_msg_id:
                        return err_msg_id
                elif node_utils.is_windows() and len(cmd_list) is 4:
                    return "OK\n\n%s" % json.dumps(
                            {"filename": cmd_list[3],
                            "path": FilePaths.NSX_FILE_STORE})
                return ("OK\n\n%s" % json.dumps(""))
            elif rc == 4:
                lg.error("Insufficient space to generate support bundle")
                return "MSG_PUBLIC_CLOUD_NO_SPACE_ERROR"
            else:
                lg.error(
                    "Error generating support-bundle, rc: %s, err: %s", rc, err)
                # Populate err_details with internal error info.
                err_details = "error=%s, details: '%s'" % (rc, err)
        except Exception as ex:
            lg.exception("Error generating support-bundle")
            # Populate err_details with python exception info.
            err_details = str(ex)
        return "MSG_PUBLIC_CLOUD_SUPPORT_ERROR\n%s" % err_details
