# ***************************************************************************
# Copyright 2014-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
from hashlib import sha256
import nsx_logging

try:
    import simplejson
except ImportError:
    import json as simplejson
import re
import socket
import ssl
import sys
import time

from cli.utils.exceptions import ApiException
from cli.utils.node_utils import is_windows
from cli.utils.error_code import ErrorCode as ec
from compat23 import compat23_httplib
from compat23.compat23_utils import compat23_urlencode

lg = nsx_logging.getLogger("cli.utils.apiclient")

SERVICE_UNAVAILABLE_MAX_RETRY_COUNT = 3

_JSESSION_COOKIE_RE = re.compile("(?:^|;| |,)JSESSIONID=([^;, ]*)")


def request(host, port, thumbprint, credentials,
            http_method, path, body=None, headers=None,
            connection_timeout=30, request_timeout=180,
            retry_count=SERVICE_UNAVAILABLE_MAX_RETRY_COUNT):
    """
    Issue an API request against an NSX API server. If a connection to
    the server cannot be established due to a connection error or a
    thumbprint mismatch, raises an ApiException.

    Args:
        host, port: Host IP and port number to send request to.
        thumbprint: The host API certificate thumbprint.
        credentials: A dictionary containing credentials to authenticate
                     with the host. Should either contain "username"
                     and "password", or "token". If all are supplied then
                     the token method will take precendence.
        http_method: "POST", "PUT", "GET", etc.
        path: Path on the host to send request to.
        body: Body of the message, if required.
        headers: Headers of the message, if required.
        connection_timeout: Raise socket.timeout if connection attempt
            exceeds this value.
        request_timeout: Raise socket.timeout if request exceeds this value.
        retry_count: Request is retried in case 'Retry-After' header is
            present in response.

    Returns:
        If a connection to the API server is established, returns a
        tuple (status, json, text, resp), where
        status is an HTTP response code
        json is the response from the server, if content-type application/json
             was returned from the API
        text is a textual description of the error, if a problem was encountered
        resp is the HTTP response object
    """
    port = int(port)
    thumbprint = thumbprint.lower()
    conn = None
    try:
        # Suppress default Python 2.7.9+ HTTPSConnection certificate and
        # hostname verification. Doing manual certificate thumbprint check
        # instead.
        if (sys.version_info >= (2, 7, 9) or
                "_create_unverified_context" in dir(ssl)):
            ctx = ssl._create_unverified_context()
            conn = compat23_httplib.HTTPSConnection(host, port, context=ctx,
                                                    timeout=connection_timeout)
        else:
            conn = compat23_httplib.HTTPSConnection(host, port,
                                                    timeout=connection_timeout)
        if is_windows():
            conn._context.check_hostname = False
        conn.connect()
        # Change conn object's timeout property to request_timeout now that
        # we have an active connection
        conn.sock.settimeout(request_timeout)
    except (socket.error, socket.timeout) as ex:
        lg.exception("%s %s raised exception: %s", http_method, path, type(ex),
                     error_code=ec.INTERNAL_ERROR)
        raise ApiException("MSG_API_CONNECTION_FAILED")
    try:
        # Verify server thumbprint
        conn_thumbprint = sha256(conn.sock.getpeercert(True)).hexdigest()
        conn_thumbprint = conn_thumbprint.lower()
        if conn_thumbprint != thumbprint:
            raise ApiException("MSG_NODE_MANAGER_BAD_THUMBPRINT")

        # Authenticate
        if "token" in credentials and credentials["token"]:
            # No need to use /session/create API. Just add authorization header with token.
            auth_hdrs = {"Authorization": "RegistrationToken " + credentials["token"]}
            auth_failed_msg = "Authentication failed - bad token"
        else:
            auth_failed_msg = "Authentication failed - bad cookie or bad xsrf token"
            # Get session cookie by calling /session/create API.
            # Setup authentication headers/cookie in the form of "JSESSIONID=..."
            login_body = compat23_urlencode(
                {"j_username": credentials.get("username"), "j_password": credentials.get("password")})
            hdrs = {"Content-Type": "application/x-www-form-urlencoded",
                    "Content-Length": len(login_body)}
            conn.request("POST", "/api/session/create", login_body, hdrs)
            resp = conn.getresponse()
            resp_body = resp.read()
            # Note: for now Manager reverse proxy returns a 302 status if the
            # authentication succeeds (btw, it also returns 302 status if
            # authentication fails... see location note below).
            if resp.status != compat23_httplib.OK:
                err_text = "Authentication failed"
                return compat23_httplib.UNAUTHORIZED, None, err_text, None
            auth_hdrs = {}
            for hdr, value in resp.getheaders():
                if hdr.lower() == "set-cookie":
                    m = _JSESSION_COOKIE_RE.search(value)
                    if m is not None:
                        auth_hdrs = {"Cookie": "JSESSIONID=%s" % m.group(1)}
                if hdr.lower() == "x-xsrf-token":
                    auth_hdrs["X-XSRF-TOKEN"] = value
            if not auth_hdrs:
                lg.debug(
                    "API authentication response missing cookie")
                err_text = "Authentication failed - missing cookie"
                return compat23_httplib.UNAUTHORIZED, None, err_text, None
            if "X-XSRF-TOKEN" not in auth_hdrs:
                lg.warning(
                    "API authentication response missing xsrf token")

        # Invoke the requested API
        headers = {} if not headers else headers
        headers.update(auth_hdrs)
        conn.request(http_method, path, body, headers)
        resp = conn.getresponse()
        content_type = resp.getheader("content-type", "").lower()
        resp_body = resp.read()
        if resp.status == compat23_httplib.UNAUTHORIZED:
            return compat23_httplib.UNAUTHORIZED, None, auth_failed_msg, None
        if resp.status not in [
            compat23_httplib.OK, compat23_httplib.CREATED, compat23_httplib.ACCEPTED,
            compat23_httplib.NO_CONTENT]:
            lg.error("%s %s returned status: %s\n%s", http_method, path,
                     resp.status, resp_body, error_code=ec.OPERATION_FAILED)
    except (socket.error, socket.timeout) as ex:
        lg.exception("%s %s raised exception: %s", http_method, path, type(ex),
                     error_code=ec.INTERNAL_ERROR)
        raise ApiException("MSG_API_CONNECTION_FAILED")
    except ApiException:
        raise
    except Exception as ex:
        lg.info("%s %s raised exception: %s", http_method, path, type(ex))
        err_text = "An unexpected error occurred."
        return compat23_httplib.INTERNAL_SERVER_ERROR, None, err_text, None
    finally:
        if conn:
            conn.close()
            conn = None

    json = text = None
    if content_type.startswith("application/json"):
        try:
            if sys.version_info >= (3,):
                resp_body = resp_body.decode("utf8")
            json = simplejson.loads(resp_body)
        except:
            err_text = "The API server returned an invalid response body."
            lg.info(err_text)
            return compat23_httplib.INTERNAL_SERVER_ERROR, None, err_text, None
    elif content_type.find("text/plain") != -1:
        if resp.status >= compat23_httplib.BAD_REQUEST:
            text = resp_body
    elif content_type.find("text/html") != -1:
        text = resp_body

    # Retry if "Retry-After" header is received in response
    if resp.getheader("Retry-After") and retry_count > 0:
        # For error codes like 429, 503, server does not process request and expect client
        # to make request again after some time. This time can be retrieved from
        # response header 'Retry-After'.
        if retry_count > SERVICE_UNAVAILABLE_MAX_RETRY_COUNT:
            retry_count = SERVICE_UNAVAILABLE_MAX_RETRY_COUNT
        retry_count = retry_count - 1
        retry_after = int(resp.getheader("Retry-After"))
        lg.info("Header Retry-After: %s, retry_count: %s", retry_after, retry_count)
        if retry_after > 60:
            retry_after = 60
        time.sleep(retry_after + 5)
        return request(host, port, thumbprint, credentials, http_method, path, body, headers,
                       connection_timeout, request_timeout, retry_count)
    return resp.status, json, text, resp
