# ***************************************************************************
# Copyright 2014-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging
import os
try:
    import simplejson
except ImportError:
    import json as simplejson
import socket
import time

from cli.utils.node_utils import get_node_type
from cli.utils.node_utils import is_appliance
from cli.utils.error_code import ErrorCode as ec
from compat23 import compat23_httplib


lg = nsx_logging.getLogger("cli.utils.rest_utils")


RETRY_INTERVAL = 1
RETRY_MAX_TIME = 15

NSXI_URI_PREFIX = "/intelligence/api/v1"
NSXI_HOST = "127.0.0.1:7447"

PROTON_URI_PREFIX = "/nsxapi/api/v1"
PROTON_HOST = "127.0.0.1:7440"

NAPI_URI_PREFIX = "/api/v1"
NAPI_HOST = "127.0.0.1:7441"


def local_node_request(http_method, path, body=None, headers=None,
                       retry_time=RETRY_MAX_TIME, retry_interval=RETRY_INTERVAL,
                       retry_if_service_unavailable=False,
                       proton_request=False, nsxi_request=False,
                       body_is_json=True):
    """Issue REST request against local /node API service (on Controllers, this
       is the NAPI service, on Managers, the Appliance Management service.

    Args:
        http_method: GET, PUT, POST, ...
        path: URI starting with /node (before submitting request path is
            prefixed with either /ws.v2 or /appliance-management/api/v1
            depending on the node type).
        body: optional body (e.g. python dictionary) to issue.
        headers: optional headers to issue (application/json is automatically
            added if body is not None).
        retry_time: retry request if failure up to retry_time seconds.
        retry_interval: wait retry_interval seconds before retry
        retry_if_service_unavailable: True, if retry request when service
            returns 503; otherwise, False.
        body_is_json: True, if request body need to be formatted in json;
            otherwise, False.
    Returns:
        (status, json, text, response) tuple where:
        status: HTTP status of request response
        json: If response is of content-type application/json, this is the
            value return from simplejson.loads(response.body)
        text: If response is text/html|plain, this the response body.  Or,
            if an error occurred, this is the error text.
        response: Response object returned from request.
    """
    host = None
    if is_appliance():
        headers = headers or {}
        if "X-NSX-Username" not in headers:
            import pwd
            headers["X-NSX-Username"] = pwd.getpwuid((os.getuid()))[0]
        if proton_request:
            host = PROTON_HOST
            path = PROTON_URI_PREFIX + path
        elif nsxi_request:
            headers["X-ROLE-ENTERPRISE_ADMIN"] = 1
            host = NSXI_HOST
            path = NSXI_URI_PREFIX + path
        else:
            host = NAPI_HOST
            path = NAPI_URI_PREFIX + path
    else:
        err = ("local_node_request invoked on unsupported node_type: %s" %
               get_node_type())
        lg.error(err, error_code=ec.OPERATION_NOT_SUPPORTED)
        return compat23_httplib.BAD_REQUEST, None, "Invalid request", None
    return rest_request(host, http_method, path, body, headers, retry_time,
                        retry_interval, retry_if_service_unavailable,
                        body_is_json)


def rest_request(host_and_port, http_method, path, body=None, headers=None,
                 retry_time=RETRY_MAX_TIME, retry_interval=RETRY_INTERVAL,
                 retry_if_service_unavailable=False, body_is_json=True):
    conn = None
    if not headers:
        headers = {}
    if body_is_json and body is not None:
        try:
            body = simplejson.dumps(body)
        except Exception:
            return compat23_httplib.BAD_REQUEST, None, "Invalid request", None
        headers["Content-Type"] = "application/json"
    start = time.time()
    resp_body = ""
    while True:
        try:
            conn = compat23_httplib.HTTPConnection(host_and_port)
            conn.request(http_method, path, body, headers)
            resp = conn.getresponse()
            content_type = resp.getheader("content-type", "").lower()
            resp_body = resp.read()
            if (resp.status == compat23_httplib.SERVICE_UNAVAILABLE and
                    retry_if_service_unavailable):
                if (time.time() - start) <= retry_time:
                    time.sleep(retry_interval)
                    continue
            break
        except socket.error as ex:
            if time.time() - start > retry_time:
                return (compat23_httplib.INTERNAL_SERVER_ERROR, None,
                        "Error connecting to the API service", None)
            time.sleep(retry_interval)
        except Exception as ex:
            lg.exception("Exception issuing %s %s" % (http_method, path),
                         error_code=ec.INTERNAL_ERROR)
            # Assume python exception will be unhelpful to the user
            # err_text = str(ex)
            err_text = "An unexpected error occurred."
            return compat23_httplib.INTERNAL_SERVER_ERROR, None, err_text, None
        finally:
            if conn:
                conn.close()

    json = text = None
    if content_type.startswith("application/json"):
        try:
            json = simplejson.loads(resp_body)
        except:
            err_text = "The API server returned an invalid response body."
            return compat23_httplib.INTERNAL_SERVER_ERROR, None, err_text, None
    elif content_type.find("text/plain") != -1:
        # NAPI providers often return webob.exc.HTTPXxx (e.g. HTTPNotFound)
        # which prepends the provider specified error text with 4 lines of
        # error summary, so if this is a text/plain error response, strip
        # the first 4 lines (note, responses from providers that use
        # compat23_httplib.Response have content-type: text/html).
        if resp.status >= compat23_httplib.BAD_REQUEST:
            if hasattr(resp_body, 'decode'):
                resp_body = resp_body.decode("utf8")
            lines = resp_body.split("\n")
            text = "\n".join(lines[4:]) if len(lines) > 4 else resp_body
        else:
            text = resp_body
    elif content_type.find("text/html") != -1:
        text = resp_body
    return resp.status, json, text, resp
