# **********************************************************************
# Copyright 2021 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
import argparse
import datetime
import logging
import os
import sys
sys.path.append("/opt/vmware/nsx-common/python/")
import time
import re
import getpass

from nsx_logging.duplicate_filter import DuplicateFilter

CONFIG_FILE_PATH = "/opt/vmware/nsx-common/python/nsx_logging/test/test_config.json"
class PythonFormatter(logging.Formatter):
    def __init__(self, name, level):
        self.name = name
        self.level = level

        self.comp = "nsx-manager"
        self.subcomp = "python"
        self.username = getpass.getuser()
        self.tags = {}

        fmt = self._structured_data_format()
        datefmt = "%Y-%m-%dT%H:%M:%S"
        super(PythonFormatter, self).__init__(fmt, datefmt)

    def format(self, record):
        record.__dict__["name"] = self.name
        record.__dict__["level"] = self.level

        record.__dict__["comp"] = self.comp
        record.__dict__["subcomp"] = self.subcomp
        record.__dict__["username"] = self.username
        record.__dict__["tags"] = self.tags

        return super(PythonFormatter, self).format(record)

    def _structured_data_format(self):
        """Construct a structured data styled format string.
        Message format that complies with RFC5424.
        """
        log_format = (
                "%(asctime)s.%(msecs)03dZ %(name)s "
                "%(level)s %(message)s")
        return log_format


class SyslogFormatter(logging.Formatter):
    def __init__(self, component, process, msgid, tags, hostname):
        self.component = component
        self.process = process
        self.msgid = msgid
        self.tags = tags
        self.hostname = hostname

        self.comp = "nsx-manager"
        self.subcomp = "python"
        self.username = getpass.getuser()

        fmt = self._structured_data_format()
        datefmt = "%Y-%m-%dT%H:%M:%S"
        super(SyslogFormatter, self).__init__(fmt, datefmt)

    def format(self, record):
        record.__dict__["component"] = self.component

        #TODO: currently, for those missing processid log msg,
        # we will give it current processid when running this script.
        # Ideally, we should leave it blank.
        if self.process != "-":
            record.__dict__["process"] = int(self.process)

        record.__dict__["msgid"] = self.msgid
        record.__dict__["hostname"] = self.hostname

        record.__dict__["comp"] =  self.tags.get("comp", self.comp)
        record.__dict__["subcomp"] = self.tags.get("subcomp", self.subcomp)
        record.__dict__["username"] = self.tags.get("username", self.username)

        if self.tags:
            record.__dict__["tags"] = "[nsx@6876"
            for k, v in self.tags.items():
                record.__dict__["tags"] += " " + k + "=\"" + v +"\""
                record.__dict__[k] = v
            record.__dict__["tags"] = record.__dict__["tags"].rstrip()
            record.__dict__["tags"] += "]"
        else:
            #for syslog library, it will print "-" for no tags log msg
            record.__dict__["tags"] = "-"
        return super(SyslogFormatter, self).format(record)

    def _structured_data_format(self):
        """Construct a structured data styled format string.
        Message format that complies with RFC5424.
        """
        log_format = (
                "%(asctime)s.%(msecs)03dZ {} %(component)s %(process)d "
                "%(msgid)s %(tags)s %(message)s")
        return log_format.format(self.hostname)

class JavaFormatter(logging.Formatter):
    def __init__(self, log_level, log_source, log_class,
            msgid,log_component, tags):
        self.log_level = log_level
        self.log_source = log_source
        self.log_class = log_class
        self.msgid = msgid
        self.log_component = log_component
        self.tags = tags

        self.comp = "nsx-manager"
        self.subcomp = "java"
        self.username = getpass.getuser()

        fmt = self._structured_data_format()
        datefmt = "%Y-%m-%dT%H:%M:%S"
        super(JavaFormatter, self).__init__(fmt, datefmt)

    def format(self, record):
        record.__dict__["log_level"] = self.log_level
        record.__dict__["log_source"] = self.log_source
        record.__dict__["log_class"] = self.log_class
        record.__dict__["msgid"] = self.msgid
        record.__dict__["log_component"] = self.log_component

        record.__dict__["comp"] =  self.tags.get("comp", self.comp)
        record.__dict__["subcomp"] = self.tags.get("subcomp", self.subcomp)
        record.__dict__["username"] = self.tags.get("username", self.username)

        if self.tags:
            record.__dict__["tags"] = "[nsx@6876"
            for k, v in self.tags.items():
                record.__dict__["tags"] += " " + k + "=\"" + v +"\""
                record.__dict__[k] = v
            record.__dict__["tags"] = record.__dict__["tags"].rstrip()
            record.__dict__["tags"] += "] "
        else:
            #for java nsx-logging library, it will print nothing for no tags log msg
            record.__dict__["tags"] = ""
        return super(JavaFormatter, self).format(record)

    def _structured_data_format(self):
        """Construct a structured data styled format string.
        Message format that complies with RFC5424.
        """
        log_format = (
                "%(asctime)s.%(msecs)03dZ %(log_level)s "
                "%(log_source)s %(log_class)s %(msgid)s %(log_component)s "
                "%(tags)s%(message)s")
        return log_format

def main():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('-i', '--input', type=str, required=True,
                    help=('path to the input log file'))
    arg_parser.add_argument('-o', '--output', type=str, required=True,
                    help=('path to the output log file'))
    arg_parser.add_argument('-t', '--type',
                    type=str, required=True,
                    help=('clarify which nsx logging library did the input'
                    'log used for emit logs, java or python or cpp(syslog)'))
    # arg_parser.add_argument('-l', '--lru_dict_size',
    #                 type=int, required=False, default=100,
    #                 help=('keep max {lru_dict_size} records in our record '
    #                 'map to do the duplicate check. Default: 100'))
    # arg_parser.add_argument('-r', '--reset_interval_in_sec',
    #                 type=int, required=False, default=10,
    #                 help=('clean the LRU dict every {reset_interval_in_sec} '
    #                 'seconds. Default: 10'))
    args = arg_parser.parse_args()

    global output_file
    global lru_dict_size
    global reset_interval_in_sec
    input_file = args.input
    output_file = args.output
    log_type = args.type
    # lru_dict_size = args.lru_dict_size
    # reset_interval_in_sec = args.reset_interval_in_sec

    if os.path.isfile(input_file) is False :
        print ("Please add parameters: \n"
            "\t-i <path to the input log file> \n"
            "\t-o <path to the output log file> \n"
            "\t-t <nsx-logging library type> \n"
            "<path to the input log file> needs to be a exist log file")
        sys.exit(1)

    if os.path.isfile(output_file):
        # user doesn't want to append, so remove the existing output file
        os.remove(output_file)

    global last_log_timestamp
    last_log_timestamp = 0
    input = open(input_file, 'r')

    #init log config
    global lg
    lg = logging.getLogger(__name__)
    lg.setLevel(logging.DEBUG)
    lg.addFilter(DuplicateFilter(config_json = CONFIG_FILE_PATH))

    # Example:
    # 2021-03-08T07:56:31.201Z manager1 NSX 25402 - [nsx@6876 comp="nsx-manager" subcomp="python" username="root" level="INFO"] test1
    if log_type == 'syslog' or log_type == 'cpp':
        last_line = ""
        for cur_line in input:
            # deal with multiple log lines in one log message case
            if not isStartWithTime(cur_line):
                if last_line == "":
                    last_line = cur_line
                else:
                    last_line += cur_line
                continue
            if last_line != "":
                # deal with the last line
                process_syslog_format_log_msg(last_line)
            last_line = cur_line
        if last_line != "":
            process_syslog_format_log_msg(last_line)
    # Example:
    # 2021-03-07T16:47:10.364Z napi.root.cluster.api_service INFO test1
    # except /var/log/nvpapi/api_roothelper.log
    elif log_type == 'python':
        last_line = ""
        for cur_line in input:
            # deal with multiple log lines in one log message case
            if not isStartWithTime(cur_line):
                if last_line == "":
                    last_line = cur_line
                else:
                    last_line += cur_line
                continue
            if last_line != "":
                # deal with the last line
                process_python_format_log_msg(last_line)
            last_line = cur_line
        if last_line != "":
            process_python_format_log_msg(last_line)
    # Example:
    # 2021-02-17T06:32:35.243Z  INFO localhost-startStop-1 MonitoringLeaderInitializationListener 14390 MONITORING [nsx@6876 comp="nsx-manager" level="INFO" subcomp="monitoring"] test1
    elif log_type == 'java':
        last_line = ""
        for cur_line in input:
            # deal with multiple log lines in one log message case
            if not isStartWithTime(cur_line):
                if last_line == "":
                    last_line = cur_line
                else:
                    last_line += cur_line
                continue
            if last_line != "":
                # deal with the last line
                process_java_format_log_msg(last_line)
            last_line = cur_line
        if last_line != "":
            process_java_format_log_msg(last_line)
    else:
        print ("We don't support " + log_type + ". Please input java/python/cpp/syslog for -t(--type)")
        input.close()
        sys.exit(1)
    input.close()


def custom_time(t):
    return time.localtime(cur_log_timestamp)

def convertTimeFromStrToInt(component, cur_time):
    if component == 'NSX':
        #2021-02-17T06:32:35.231Z
        try:
            time_tuple = datetime.datetime.strptime(cur_time, "%Y-%m-%dT%H:%M:%S.%fZ").timetuple()
        except ValueError:
            time_tuple = datetime.datetime.strptime(cur_time, "%Y-%m-%d %H:%M:%S,%f").timetuple()
        return int(time.mktime(time_tuple))
    else:
        #2021-02-27T14:26:02.352257+00:00
        new_cur_time = cur_time[0: -3] + cur_time[-2:]
        time_tuple = datetime.datetime.strptime(new_cur_time, "%Y-%m-%dT%H:%M:%S.%f%z").timetuple()
        return int(time.mktime(time_tuple))

def isStartWithTime(line):
    try:
        time_tuple = datetime.datetime.strptime(line[0: 19], "%Y-%m-%dT%H:%M:%S").timetuple()
    except ValueError:
        return False
    return True

def process_python_format_log_msg(line):
    # 2021-03-07T16:47:10.364Z napi.root.cluster.api_service INFO No change in api_service global config
    global cur_log_timestamp
    global last_log_timestamp
    if line.startswith("<182>1"):
        line = line[7:]
    line = re.split(r"[ ]+", line)
    log_time = line[0]
    name = line[1]
    level = line[2]
    content = ' '.join(line[3:]).strip('\n')

    fh = logging.FileHandler(output_file)
    fomatter = PythonFormatter(name=name, level=level)

    #get the log time from input log
    cur_log_timestamp = convertTimeFromStrToInt('NSX', log_time)
    if last_log_timestamp != 0 and cur_log_timestamp > last_log_timestamp:
        time.sleep(cur_log_timestamp - last_log_timestamp)
    last_log_timestamp = cur_log_timestamp
    fomatter.converter = custom_time

    fh.setFormatter(fomatter)
    lg.addHandler(fh)

    #try to log the contents
    lg.info(content)
    lg.removeHandler(fh)

def process_syslog_format_log_msg(line):
    global cur_log_timestamp
    global last_log_timestamp
    if line.startswith("<182>1"):
        line = line[7:]
    line = re.split(r"[ ]+", line)
    log_time = line[0]
    hostname = line[1]
    component = line[2]
    process = line[3]
    msgid = line[4]
    full_content = ' '.join(line[5:])

    # get the tag_dict
    tag_dict = {}
    # check the content inside [] which start with nsx@6876
    if re.findall(r'\[(nsx@6876.*?)\]', full_content):
        tag_list = re.findall(r'\[(nsx@6876.*?)\]', full_content)[0].split(" ")
        for tag in tag_list:
            details = tag.split("=")
            if len(details) == 2:
                tag_dict[details[0]] = details[1][1:-1]

    # get the content
    if tag_dict:
        pos = full_content.index("]") + 2
        content = full_content[pos:].strip('\n')
    else:
        content = full_content[2:].strip('\n')

    fh = logging.FileHandler(output_file)

    fomatter = SyslogFormatter(component=component,
            process=process,msgid=msgid,tags=tag_dict, hostname=hostname)

    #get the log time from input log
    cur_log_timestamp = convertTimeFromStrToInt(component, log_time)
    if last_log_timestamp != 0 and cur_log_timestamp > last_log_timestamp:
        time.sleep(cur_log_timestamp - last_log_timestamp)
    last_log_timestamp = cur_log_timestamp
    fomatter.converter = custom_time

    fh.setFormatter(fomatter)
    lg.addHandler(fh)
    adapter = logging.LoggerAdapter(lg, tag_dict)

    #try to log the contents
    adapter.info(content)
    lg.removeHandler(fh)

def process_java_format_log_msg(line):
    global cur_log_timestamp
    global last_log_timestamp
    line = re.split(r"[ ]+", line)
    log_time = line[0]
    log_level = line[1]
    log_source = line[2]
    log_class = line[3]
    msgid = line[4]
    log_component = line[5]
    full_content = ' '.join(line[6:])

    # get the tag_dict
    tag_dict = {}
    # check the content inside [] which start with nsx@6876
    if re.findall(r'\[(nsx@6876.*?)\]', full_content):
        tag_list = re.findall(r'\[(nsx@6876.*?)\]', full_content)[0].split(" ")
        for tag in tag_list:
            details = tag.split("=")
            if len(details) == 2:
                tag_dict[details[0]] = details[1][1:-1]

    # get the content
    if tag_dict:
        pos = full_content.index("]") + 2
        content = full_content[pos: ].strip('\n')
    else:
        log_component = ""
        full_content = ' '.join(line[5:])
        content = full_content.strip('\n')

    fh = logging.FileHandler(output_file)
    fomatter = JavaFormatter(log_level=log_level, log_source=log_source,
            log_class=log_class,msgid=msgid,log_component=log_component, tags=tag_dict)

    # get the log time from input log
    cur_log_timestamp = convertTimeFromStrToInt('NSX', log_time)
    if last_log_timestamp != 0 and cur_log_timestamp > last_log_timestamp:
        time.sleep(cur_log_timestamp - last_log_timestamp)
    last_log_timestamp = cur_log_timestamp
    fomatter.converter = custom_time

    fh.setFormatter(fomatter)
    lg.addHandler(fh)
    adapter = logging.LoggerAdapter(lg, tag_dict)

    #try to log the contents
    adapter.info(content)
    lg.removeHandler(fh)

if __name__ == "__main__":
    main()