# Copyright (C) 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
import os
import socket
import struct
import sys


def create_socket():
    if sys.platform == 'win32':
        sock_af = socket.AF_INET
    else:
        sock_af = socket.AF_UNIX

    sock = socket.socket(sock_af, socket.SOCK_STREAM)
    return sock


def create_and_bind_socket(sock_file, wait_time=0):
    sock = create_socket()
    if sys.platform == 'win32':
        # For win32, sock_file is in the form of (IP address, port number)
        sock.bind(sock_file)
    else:
        # Create directory for socket file
        try:
            dir_name = os.path.dirname(sock_file)
            os.makedirs(dir_name)
        except OSError:
            pass
        # Delete the existing socket file if any
        try:
            os.remove(sock_file)
        except OSError:
            pass
        sock.bind(sock_file)
        # On ESXi, no file will be created
        if os.path.exists(sock_file):
            os.chmod(sock_file, 0o640)
    if wait_time > 0:
        sock.listen(wait_time)
    return sock


def connect_socket(sock_file):
    sock = create_socket()
    sock.connect(sock_file)
    return sock


def close_socket(sock):
    if sock:
        sock.close()


def accept_conn(sock):
    return sock.accept()[0]


def close_conn(conn):
    if conn:
        conn.close()


def send_msg(sock, msg):
    msg = msg.encode()
    msg_len = len(msg)
    header = struct.pack("<I", msg_len)
    sent = sock.send(header)
    if sent != 4:
        return False, "Failed to send message header"
    total_sent = 0
    while total_sent < msg_len:
        sent = sock.send(msg[total_sent:])
        if sent == 0:
            return False, "Socket connection is broken"
        total_sent += sent
    return True, ""


def receive_msg(sock):
    header = sock.recv(4)
    if len(header) != 4:
        return None, "Failed to receive message header"
    msg_len = struct.unpack("<I", header)[0]
    msg = bytes()
    total_received = 0
    while total_received < msg_len:
        data = sock.recv(msg_len - total_received)
        if not data:
            return None, "Socket connection is broken"
        msg += data
        total_received += len(data)
    return msg.decode(), ""
