# Copyright (C) 2019-2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import os

import socket
import traceback
import sha.config
import sha.preconfig
import yaml
from sha.contrib.metric.abc_base import DedicateThreadMetric
from sha.contrib.metric.utils import timeout_cache

from sha.config import SHA_ROOT_DIR
from vmware.nsx.daemon import health_pb2
from vmware.nsx.rpc import NsxRpcCallError
from vmware.nsx.rpc import NsxRpcClient
from vmware.nsx.sha import sha_pb2

_logger = sha.preconfig.get_sub_logger(__name__, s2comp="metric-nsx.daemon.daemon-health")

DEFAULT_TIMEOUT = 2
# default config - treat daemon is down only when the rpc echo failed TOLERANCE_COUNT+1 times continuously.
DEFAULT_OPSAGENT_TOLERANCE_COUNT = 2
DEFAULT_NESTDB_TOLERANCE_COUNT = 3
DEFAULT_CFGAGENT_TOLERANCE_COUNT = 0
DEFAULT_NSXAGENT_TOLERANCE_COUNT = 0

# read tolerance config from file
sha_config_path = os.path.join(SHA_ROOT_DIR, 'sha_config.yml')
try:
    with open(sha_config_path, 'r') as f:
        sha_config = yaml.load(f.read(), Loader=yaml.Loader)
        OPSAGENT_TOLERANCE_COUNT = int(sha_config.get('OPSAGENT_TOLERANCE_COUNT', DEFAULT_OPSAGENT_TOLERANCE_COUNT))
        NESTDB_TOLERANCE_COUNT = int(sha_config.get('NESTDB_TOLERANCE_COUNT', DEFAULT_NESTDB_TOLERANCE_COUNT))
        CFGAGENT_TOLERANCE_COUNT = int(sha_config.get('CFGAGENT_TOLERANCE_COUNT', DEFAULT_CFGAGENT_TOLERANCE_COUNT))
        NSXAGENT_TOLERANCE_COUNT = int(sha_config.get('NSXAGENT_TOLERANCE_COUNT', DEFAULT_NSXAGENT_TOLERANCE_COUNT))
except Exception as e:
    trace = traceback.format_exc()
    _logger.warning('Failed to read tolerance config from sha_config.yml, will use default one. exception:{}, trace:{}'
                    .format(e, trace))
    OPSAGENT_TOLERANCE_COUNT = DEFAULT_OPSAGENT_TOLERANCE_COUNT
    NESTDB_TOLERANCE_COUNT = DEFAULT_NESTDB_TOLERANCE_COUNT
    CFGAGENT_TOLERANCE_COUNT = DEFAULT_CFGAGENT_TOLERANCE_COUNT
    NSXAGENT_TOLERANCE_COUNT = DEFAULT_NSXAGENT_TOLERANCE_COUNT

_logger.info('OPSAGENT_TOLERANCE_COUNT: {}, NESTDB_TOLERANCE_COUNT: {}, CFGAGENT_TOLERANCE_COUNT: {}, '
             'NSXAGENT_TOLERANCE_COUNT: {}'.format(OPSAGENT_TOLERANCE_COUNT, NESTDB_TOLERANCE_COUNT,
                                                   CFGAGENT_TOLERANCE_COUNT, NSXAGENT_TOLERANCE_COUNT))

DAEMON_TOLERANCE_COUNT_SETTING = {
    sha_pb2.AgentStatusMsg.AgentType.Value('NSX_OPSAGENT'): OPSAGENT_TOLERANCE_COUNT,
    sha_pb2.AgentStatusMsg.AgentType.Value('NSX_CFGAGENT'): CFGAGENT_TOLERANCE_COUNT,
    sha_pb2.AgentStatusMsg.AgentType.Value('NSX_NESTDB'): NESTDB_TOLERANCE_COUNT,
    sha_pb2.AgentStatusMsg.AgentType.Value('NSX_AGENT_DAEMON'): NSXAGENT_TOLERANCE_COUNT
}

# endpoint
if sha.config.SYSTEM == sha.config.PLATFORM_LINUX:
    OPSAGENT_DEFAULT_ENDPOINT = "unix:///var/run/vmware/nsx-opsagent/opsagent-dh.sock"
    NESTDB_DEFAULT_ENDPOINT = "unix:///var/run/vmware/nestdb/nestdb-server.sock"
    THREATDETECTOR_DEFAULT_ENDPOINT = "unix:///var/run/vmware/nsx-agent/threat-detector.sock"
    NSXAGENT_DEFAULT_ENDPOINT = "unix:///var/run/vmware/nsx-agent/nsx-agent-dh.sock"
else:  # system == 'Windows' || system == 'VMkernel'
    OPSAGENT_DEFAULT_ENDPOINT = "tcp://127.0.0.1:4098"
    NESTDB_DEFAULT_ENDPOINT = "tcp://127.0.0.1:2480"
    THREATDETECTOR_DEFAULT_ENDPOINT = "tcp://127.0.0.1:4099"
    NSXAGENT_DEFAULT_ENDPOINT = "tcp://127.0.0.1:4097"

# only for system == 'VMkernel'
CFGAGENT_DEFAULT_ENDPOINT = "tcp://127.0.0.1:4097"

daemon_default_endpoint = {
    sha_pb2.AgentStatusMsg.AgentType.Value('NSX_OPSAGENT'): OPSAGENT_DEFAULT_ENDPOINT,
    sha_pb2.AgentStatusMsg.AgentType.Value('NSX_CFGAGENT'): CFGAGENT_DEFAULT_ENDPOINT,
    sha_pb2.AgentStatusMsg.AgentType.Value('NSX_NESTDB'): NESTDB_DEFAULT_ENDPOINT,
    sha_pb2.AgentStatusMsg.AgentType.Value('NSX_AGENT_DAEMON'): NSXAGENT_DEFAULT_ENDPOINT
}

# Build global agent state list
if sha.config.SYSTEM == sha.config.PLATFORM_ESX:
    daemon_agent_state = {
        sha_pb2.AgentStatusMsg.AgentType.Value('NSX_OPSAGENT'): sha_pb2.AgentStatus.Value('DOWN'),
        sha_pb2.AgentStatusMsg.AgentType.Value('NSX_NESTDB'): sha_pb2.AgentStatus.Value('DOWN'),
        sha_pb2.AgentStatusMsg.AgentType.Value('NSX_CFGAGENT'): sha_pb2.AgentStatus.Value('DOWN'),
    }
elif sha.config.SYSTEM == sha.config.PLATFORM_LINUX:
    daemon_agent_state = {
        sha_pb2.AgentStatusMsg.AgentType.Value('NSX_OPSAGENT'): sha_pb2.AgentStatus.Value('DOWN'),
        sha_pb2.AgentStatusMsg.AgentType.Value('NSX_NESTDB'): sha_pb2.AgentStatus.Value('DOWN'),
        sha_pb2.AgentStatusMsg.AgentType.Value('NSX_AGENT_DAEMON'): sha_pb2.AgentStatus.Value('DOWN')
    }
else:  # system == Windows
    daemon_agent_state = {
        sha_pb2.AgentStatusMsg.AgentType.Value('NSX_OPSAGENT'): sha_pb2.AgentStatus.Value('DOWN'),
        sha_pb2.AgentStatusMsg.AgentType.Value('NSX_NESTDB'): sha_pb2.AgentStatus.Value('DOWN'),
        sha_pb2.AgentStatusMsg.AgentType.Value('NSX_AGENT_DAEMON'): sha_pb2.AgentStatus.Value('DOWN')
    }


def daemon_health_rpc_client(stub=None, endpoint=None):
    try:
        return NsxRpcClient(stub, endpoint)
    except (socket.timeout, socket.error) as e:
        trace = traceback.format_exc()
        _logger.warning('Failed to create RPC Client - (endpoint {}), exception:{}, trace:{}'
                        .format(endpoint, e, trace))
        return None


def agent_name(agent_type):
    if agent_type == sha_pb2.AgentStatusMsg.AgentType.Value('NSX_OPSAGENT'):
        return 'nsx-opsAgent'
    elif agent_type == sha_pb2.AgentStatusMsg.AgentType.Value('NSX_NESTDB'):
        return 'nsx-nestdb'
    elif agent_type == sha_pb2.AgentStatusMsg.AgentType.Value('NSX_CFGAGENT'):
        return 'nsx-cfgAgent'
    elif agent_type == sha_pb2.AgentStatusMsg.AgentType.Value('NSX_AGENT_DAEMON'):
        return 'nsx-agent'
    else:
        return 'invalid agent'


def build_agent_state():
    agent_state = sha_pb2.TnAgentStateMsg()
    down_count = 0

    for agent_type in daemon_agent_state.keys():
        if daemon_agent_state[agent_type] == sha_pb2.AgentStatus.Value('DOWN'):
            down_count += 1
        agent_status_msg = agent_state.agent_status.add()
        agent_status_msg.type = agent_type
        agent_status_msg.status = daemon_agent_state[agent_type]

    if down_count != 0:
        agent_state.total_status = sha_pb2.AgentStatus.Value('DOWN')
    else:
        agent_state.total_status = sha_pb2.AgentStatus.Value('UP')

    agent_state.up_count = len(daemon_agent_state) - down_count
    agent_state.down_count = down_count

    _logger.debug("agent state: {}".format(agent_state))

    return agent_state


def build_threat_state(ts_state):
    if ts_state is None:
        return sha_pb2.TnThreatStateMsg()

    assert isinstance(ts_state, sha_pb2.TnThreatStateMsg)
    tn_threat_state = ts_state
    if daemon_agent_state[sha_pb2.AgentStatusMsg.AgentType.Value('NSX_NESTDB')] == sha_pb2.AgentStatus.Value('DOWN'):
        tn_threat_state.state = sha_pb2.TnThreatStateMsg.TnState.Value(
            'TN_STATE_UNKNOWN')

    _logger.debug("threat state: %s", tn_threat_state)
    return tn_threat_state


class DaemonHealth(DedicateThreadMetric):

    def __init__(self):
        self._logger = sha.config.get_logger(__name__)
        self.daemon_rpc_client_map = {}
        self.threatstate_rpc_client = None
        self._daemon_tolerance_count = DAEMON_TOLERANCE_COUNT_SETTING.copy()

    @timeout_cache(2)
    def usage(self, *args, **kwargs):
        self.populate_each_daemon_rpc_client()
        self.update_daemon_agent_state()

        self.populate_threatstate_rpc_client()
        ts_state = self.get_threat_state()

        tn_agent_status = sha_pb2.TnAgentStatusMsg()
        tn_agent_status.threat_state.CopyFrom(build_threat_state(ts_state))
        tn_agent_status.agent_state.CopyFrom(build_agent_state())
        return tn_agent_status

    def clean(self):
        _logger.info("start cleaning all rpc clients.")
        # close RPC clients
        for rpc_key in self.daemon_rpc_client_map:
            rpc_client = self.daemon_rpc_client_map[rpc_key]
            if rpc_client is not None:
                try:
                    rpc_client.Destroy()
                    _logger.info("Succeed in destroying rpc client:{}".format(agent_name(rpc_key)))
                except Exception:
                    trace = traceback.format_exc()
                    _logger.error(
                        "Failed in destroying rpc client:{}, trace:{}".format(agent_name(rpc_key), trace))

        if self.threatstate_rpc_client is not None:
            try:
                self.threatstate_rpc_client.Destroy()
                _logger.info("Succeed in destroying threat state rpc client")
            except Exception:
                trace = traceback.format_exc()
                _logger.error("Failed in destroying threat state rpc client, trace:{}".format(trace))

    def update_daemon_agent_state(self):
        for agent_type in self.daemon_rpc_client_map:
            self.echo('ping', agent_type)

    def populate_threatstate_rpc_client(self):
        if sha.config.NODE_TYPE != sha.config.DIST_BMS:
            return
        if self.threatstate_rpc_client is None:
            _logger.debug('RPC connecting to threat state server')
            self.threatstate_rpc_client = \
                daemon_health_rpc_client(sha_pb2.ThreatStateService_Stub, THREATDETECTOR_DEFAULT_ENDPOINT)

    def populate_each_daemon_rpc_client(self):
        for agent_type in daemon_agent_state.keys():
            if agent_type not in self.daemon_rpc_client_map or self.daemon_rpc_client_map[agent_type] is None:
                _logger.debug('RPC connecting to Agent: {}'.format(agent_name(agent_type)))
                self.daemon_rpc_client_map[agent_type] = daemon_health_rpc_client(health_pb2.DaemonHealth_Stub,
                                                                                  daemon_default_endpoint[agent_type])

    def get_threat_state(self):
        # threat-detector only works on BMS now!
        if sha.config.NODE_TYPE != sha.config.DIST_BMS:
            return None

        res = sha_pb2.TnThreatStateMsg()
        res.state = sha_pb2.TnThreatStateMsg.TnState.Value('TN_STATE_UNKNOWN')
        if self.threatstate_rpc_client is None:
            return res

        request = sha_pb2.TnThreatStateRequestMsg()
        try:
            response = self.threatstate_rpc_client.GetThreatState(request)

            if not isinstance(response, sha_pb2.TnThreatStateMsg):
                _logger.warning("threat state return wrong response: {}".format(response))
                return res

            res = response
            _logger.debug("threat state return response: {}".format(res))
        except NsxRpcCallError as e:
            trace = traceback.format_exc()
            _logger.info("Error in checking threat state, error: {}, trace:{}".format(e, trace))
            try:
                self.threatstate_rpc_client.Destroy()
                self.threatstate_rpc_client = None
                _logger.info("Succeed in destroying threat state rpc client")
            except Exception:
                trace = traceback.format_exc()
                _logger.error("Failed in destroying threat state rpc client, trace:{}".format(trace))

        return res

    def echo(self, echo_msg, agent_type):
        daemon_agent_state[agent_type] = sha_pb2.AgentStatus.Value('DOWN')

        daemon_rpc_client = self.daemon_rpc_client_map[agent_type]
        if daemon_rpc_client is None:
            self.reduce_tolerance(agent_type)
            return

        req = health_pb2.EchoMsg()
        req.msg = echo_msg
        try:
            rsp = daemon_rpc_client.Echo(req, DEFAULT_TIMEOUT)
            if not isinstance(rsp, health_pb2.EchoMsg):
                _logger.warning("Daemon:{} response is not a health_pb2.EchoMsg, rsp:{}"
                                .format(agent_name(agent_type), rsp))
                self.reduce_tolerance(agent_type)
                return

            daemon_agent_state[agent_type] = sha_pb2.AgentStatus.Value('UP')
            self.restore_tolerance(agent_type)

            _logger.debug("echo Daemon:{} response: {}".format(agent_name(agent_type), rsp))
        except NsxRpcCallError as e:
            trace = traceback.format_exc()
            _logger.info(
                "Error in echo Daemon:{}, error: {}, trace:{}".format(agent_name(agent_type), e, trace))

            self.reduce_tolerance(agent_type)

            try:
                self.daemon_rpc_client_map[agent_type].Destroy()
                self.daemon_rpc_client_map[agent_type] = None
                _logger.info("Succeed in destroying rpc client:{}".format(agent_name(agent_type)))
            except Exception:
                trace = traceback.format_exc()
                _logger.error(
                    "Failed in destroying rpc client:{}, trace:{}".format(agent_name(agent_type), trace))

    def reduce_tolerance(self, agent_type):
        """
        Treat agent is UP as long as its current tolerance > 0.
        :param agent_type:
        :return:
        """
        current_count = self._daemon_tolerance_count[agent_type]
        if current_count > 0:
            daemon_agent_state[agent_type] = sha_pb2.AgentStatus.Value('UP')
            self._daemon_tolerance_count[agent_type] = current_count - 1
            _logger.info(
                "{} current tolerance:{}".format(agent_name(agent_type), self._daemon_tolerance_count[agent_type]))

    def restore_tolerance(self, agent_type):
        self._daemon_tolerance_count[agent_type] = DAEMON_TOLERANCE_COUNT_SETTING[agent_type]
