# Copyright (C) 2019-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
__all__ = ['ShaNestdbClient']

import random
import threading
import time
import traceback

from vmware.nsx.nestdb import nestdb_common_pb2 as nestdb_pb2
from vmware.nsx.nestdb import nestdb_ops_upm_pb2 as ops_pb2
from vmware.nsx.nestdb import nestdb_schema_upm_pb2 as schema_pb2
from vmware.nsx.nestdb import nestdb_ops_common_pb2 as ops_common_pb2
from vmware.nsx.nestdb import nestdb_schema_common_pb2 as schema_common_pb2
from vmware.nsx.nestdb import nestdb_types_pb2 as types_pb2
from vmware.nsx.nestdb.nestdb_pb2 import NestDb_Stub, MonitorRequestMsg

import sha.preconfig
from sha.core.profile._resolver import Resolver
from sha.core.profile._global_config_resolver import GlobalConfigResolver
from sha.core.profile._host_config_resolver import HostConfigResolver
from sha.core.profile.custom_plugin_resolver import CustomPluginResolver
from vmware.nsx.rpc import NsxRpcClient
from vmware.nsx.rpc.status import CallStatus
from sha.utils import convert_to_uuid_str, sleep_and_check_running

if sha.config.SYSTEM == sha.config.PLATFORM_LINUX:
    NESTDB_ADDR = "unix:///var/run/vmware/nestdb/nestdb-server.sock"
else:  # system == 'Windows' || system == 'VMkernel'
    NESTDB_ADDR = "tcp://127.0.0.1:2480"

_logger = sha.preconfig.get_sub_logger(__name__, 'nestdb-client')


def is_clear(op_msg):
    return op_msg.self.op == nestdb_pb2.OpMsg.Operation.Value('CLEAR')


def get_profile_type(name):
    return schema_pb2.ProfileType.DESCRIPTOR.values_by_name[name].number


def get_profile_type_name(value):
    return schema_pb2.ProfileType.DESCRIPTOR.values_by_number[value].name


def get_object_type_name(value):
    return types_pb2._OBJECTTYPE.values_by_number[value].name


def should_handle(profile_msg, old_profile_msg):
    sha_cared_types = {
        get_profile_type('SYSTEM_HEALTH_PLUGIN'),
        get_profile_type('SYSTEM_HEALTH'),
        get_profile_type('COMMON_AGENT_HOST_CONFIG'),
    }
    if profile_msg.profile_type in sha_cared_types:
        return True
    if old_profile_msg.profile_type in sha_cared_types:
        return True
    return False


class MsgCache:
    """
    Cached all nestdb msg objs used by SHA system currently, they are:
    1. SHA profiles
    2. custom plugin profiles
    3. Global config
    """

    current_objs = {}

    @classmethod
    def update_with_delta_msg(cls, obj):
        """
        Update/Delete msg in cache accordingly
        :param msg:
        :return:
        """
        m_id, op_msg = cls.get_msg_id_and_op_msg(obj)

        if m_id is None:
            # SHA do not care this msg
            return False

        if is_clear(op_msg):
            cls.current_objs.pop(m_id)
            _logger.info("handle_delta_msg - remove msg. msg_id:{}, current cache msg_ids:{}"
                         .format(m_id, cls.current_objs.keys()))
        else:
            cls.current_objs[m_id] = obj
            _logger.info("handle_delta_msg - update msg. msg_id:{}, current cache msg_ids:{}"
                         .format(m_id, cls.current_objs.keys()))
        return True

    @classmethod
    def update_with_full_msg(cls, fullsync_objs):
        """
        Update cache with all msgs in full sync;
        For those msgs not in fullsync, remove them from cache. At the same time, create 'CLEAR' msgs to remove them
        from system.

        Return a msg list for all msgs in cache and those 'CLEAR' msgs.
        """
        new_objs = {}
        for obj in fullsync_objs:
            m_id, _ = cls.get_msg_id_and_op_msg(obj)
            if m_id is None:
                # SHA do not care this msg
                continue
            new_objs[m_id] = obj

        # Update cache with all msgs in full sync;
        cls.current_objs.update(new_objs)
        # For those msgs not in fullsync, remove them from cache. At the same time, create 'CLEAR' msgs to remove them
        # from system.
        res = []
        all_msg_ids = list(cls.current_objs.keys())
        for cur_msg_id in all_msg_ids:
            if cur_msg_id in new_objs:
                res.append(cls.current_objs[cur_msg_id])
                _logger.info("handle_fullsync_msg - SET msg. msg_id:{}".format(cur_msg_id))
                continue
            # remove them from cache
            obj_to_remove = cls.current_objs.pop(cur_msg_id)
            # convert to clear msg
            obj_clear = cls.convert_to_clear_msg(obj_to_remove)
            res.append(obj_clear)
            _logger.info("handle_fullsync_msg - CLEAR msg. msg_id:{}".format(cur_msg_id))

        _logger.info("handle_fullsync_msg - after handle, current cache msg_ids:{}".format(cls.current_objs.keys()))
        return res

    @classmethod
    def convert_to_clear_msg(cls, obj):
        """
        Convert the input 'SET' msg to a 'CLEAR' msg
        :param obj: 'SET' msg
        :return: 'CLEAR' msg
        """
        obj.old_object_body = obj.value
        if obj.object_type == types_pb2.PROFILE:
            op_msg = ops_pb2.ProfileOpMsg()
            op_msg.ParseFromString(obj.op_msg)
            op_msg.self.op = nestdb_pb2.OpMsg.Operation.Value('CLEAR')
        else:
            op_msg = ops_common_pb2.GlobalConfigOpMsg()
            op_msg.ParseFromString(obj.op_msg)
            op_msg.self.op = nestdb_pb2.OpMsg.Operation.Value('CLEAR')
        obj.op_msg = op_msg.SerializeToString()
        return obj

    @classmethod
    def get_msg_id_and_op_msg(cls, obj):
        """
        parse input msg to get its id and op_msg.
        For PROFILE msg, if SHA do not handle it, return None.
        """
        if not obj.op_msg and not obj.value and not obj.old_object_body:
            # this only occurs in the full sync case, if nestdb does not have any object for the monitored
            # object type
            _logger.info("get_msg_id_and_op_msg - obj is empty, no need to handle this. "
                         "object_type name:{}".format(get_object_type_name(obj.object_type)))
            return None, None

        if obj.object_type == types_pb2.PROFILE:
            msg = schema_pb2.ProfileMsg()
            msg.ParseFromString(obj.value)
            old_msg = schema_pb2.ProfileMsg()
            old_msg.ParseFromString(obj.old_object_body)
            if not should_handle(msg, old_msg):
                _logger.info("get_msg_id_and_op_msg - Not a SHA profile type, ignore it. msg_id:{}. "
                             "msg.profile_type name:{}, old_msg.profile_type name:{}, "
                             .format(convert_to_uuid_str(msg.id), get_profile_type_name(msg.profile_type),
                                     get_profile_type_name(old_msg.profile_type)))
                return None, None
            op_msg = ops_pb2.ProfileOpMsg()
            op_msg.ParseFromString(obj.op_msg)
        else:  # obj.object_type == types_pb2.GLOBAL_CONFIG
            msg = schema_common_pb2.GlobalConfigMsg()
            msg.ParseFromString(obj.value)
            op_msg = ops_common_pb2.GlobalConfigOpMsg()
            op_msg.ParseFromString(obj.op_msg)

        uuid_msg = msg.id
        msg_id = convert_to_uuid_str(uuid_msg)
        _logger.info("get_msg_id_and_op_msg - Parsed msg from nestdb. object_type:{}, msg_id:{}, op_msg:{}, msg:{}, "
                     .format(obj.object_type, msg_id, op_msg, msg))

        return msg_id, op_msg


class ShaNestdbClient(threading.Thread):

    def __init__(self, name, *args, **kwargs):
        super().__init__(name='nsx-sha-nestdb-client', *args, **kwargs)
        _logger.info("nestdb client initializing")
        self._name = name
        self._endpoint = NESTDB_ADDR
        self._stub = None
        self._keep_running = True
        # The first update is a full sync for nestdb current state, when sha setup monitor with nestdb;
        # This indicates whether the full sync transaction is done
        self._full_sync_transaction_done = False
        # whether the should re-setup monitor to nestdb
        self._remonitor = False
        # A transaction may span multi-UpdateMsgs from nestdb, this stores all msg objs for a full sync transaction.
        self._full_sync_msg_objs = []

        # for monitor request/response
        self._monitored_object_types = []
        self._client_context = None

        # TODO for mock test
        # CustomPluginResolver()

    # noinspection PyBroadException
    def run(self):
        _logger.info('ShaNestdbClient is starting...')
        if self._setup_monitor():
            while self._keep_running:
                # this thread must not terminate, otherwise it cannot receive the callback from nestdb.
                try:
                    if self._remonitor:
                        if not self._setup_monitor():
                            break
                    sleep_and_check_running(self, 1)
                except Exception as e:
                    trace = traceback.format_exc()
                    _logger.warning("Exception in ShaNestdbClient running, exception:{}, trace:{}".format(e, trace))

        self._destroy_rpc_stub()
        _logger.info('ShaNestdbClient is stopped.')

    def _setup_monitor(self):
        """
        Setup monitor on nestdb server
        """
        while self._keep_running:
            try:
                # remove previous stale stub
                if self._stub is not None:
                    _logger.info('stub is not None, destroy it first before setup monitor.')
                    self._destroy_rpc_stub()

                self._full_sync_transaction_done = False
                self._full_sync_msg_objs.clear()

                # setup connection
                self._stub = NsxRpcClient(NestDb_Stub, self._endpoint)
                self._client_context = self._stub.CreateClientContext()
                self._client_context.SetClientId(self._name)
                _logger.info('connect nestdb successfully.')

                # setup monitor
                self._monitor()
                self._remonitor = False
                _logger.info('started monitoring nestdb')

                return True
            except Exception as e:
                trace = traceback.format_exc()
                _logger.warning("Exception in connecting nestdb, "
                                "will retry in 10s. Exception:{}, trace:{}".format(e, trace))
                sleep_and_check_running(self, 10)

        _logger.warning("Failed in setting up monitor on nestdb.")
        return False

    def _monitor(self):
        """
        Send monitor request to nestdb server.
        This function should only be invoked after all monitor settings are
        done, and can only be invoked once.
        """
        request_msg = MonitorRequestMsg()

        for object_type in [types_pb2.PROFILE, types_pb2.GLOBAL_CONFIG]:
            msg = request_msg.statement.add()
            msg.include_old_values = True
            msg.object.object_type = object_type

        request_msg.monitor_request_id = "sha_monitor_req_id_" + str(random.randint(1, 100))
        _logger.debug(request_msg)
        self._stub.Monitor_async(request_msg, self._monitor_response_handler, self._client_context)

    def _handle_global_config_msg(self, obj):
        msg = schema_common_pb2.GlobalConfigMsg()
        msg.ParseFromString(obj.value)
        _logger.info("Global config message:{}, msg_id:{}".format(msg, convert_to_uuid_str(msg.id)))
        op_msg = ops_common_pb2.GlobalConfigOpMsg()
        op_msg.ParseFromString(obj.op_msg)
        _logger.info("Global config op message:{}".format(op_msg))
        old_msg = None
        if obj.old_object_body is not None:
            old_msg = schema_common_pb2.GlobalConfigMsg()
            old_msg.ParseFromString(obj.old_object_body)
        if is_clear(op_msg):
            if old_msg is not None:
                GlobalConfigResolver.clear(old_msg)
        else:  # SET
            GlobalConfigResolver.set(msg, old_msg)

    def _handle_profile_msg(self, obj):
        profile_msg = schema_pb2.ProfileMsg()
        profile_msg.ParseFromString(obj.value)
        profile_op_msg = ops_pb2.ProfileOpMsg()
        profile_op_msg.ParseFromString(obj.op_msg)
        _logger.info("_handle_profile_msg - profile_msg_id:{}, profile_op_msg: {}, profile_msg:{}"
                     .format(convert_to_uuid_str(profile_msg.id), profile_op_msg, profile_msg))

        if is_clear(profile_op_msg):
            if obj.old_object_body is not None:
                old_profile_msg = schema_pb2.ProfileMsg()
                old_profile_msg.ParseFromString(obj.old_object_body)
                if old_profile_msg.profile_type == get_profile_type("SYSTEM_HEALTH_PLUGIN"):
                    _logger.info("_handle_profile_msg - remove SYSTEM_HEALTH_PLUGIN:{}".format(
                        old_profile_msg.system_health_plugin_profile.name))
                    CustomPluginResolver.resolve(old_profile_msg, True)
                elif old_profile_msg.profile_type == get_profile_type("SYSTEM_HEALTH"):
                    Resolver.clear(old_profile_msg)
                else: # COMMON_AGENT_HOST_CONFIG
                    HostConfigResolver.clear(old_profile_msg.common_agent_host_config)
        else:
            if profile_msg.profile_type == get_profile_type("SYSTEM_HEALTH_PLUGIN"):
                _logger.info(
                    "_handle_profile_msg - add SYSTEM_HEALTH_PLUGIN:{}".format(
                        profile_msg.system_health_plugin_profile.name))
                CustomPluginResolver.resolve(profile_msg, False)
            elif profile_msg.profile_type == get_profile_type("SYSTEM_HEALTH"):
                Resolver.set(profile_msg)
            else: # COMMON_AGENT_HOST_CONFIG
                HostConfigResolver.set(profile_msg.common_agent_host_config)

    def _monitor_response_handler(self, message, status):
        try:
            if status is not None:
                # some error happened
                _logger.error("_monitor_response_handler - Failed to monitor, status: {}".format(status))
                if status.code in {CallStatus.COMMUNICATION_ERROR, CallStatus.INTERNAL} and self._keep_running:
                    self._remonitor = True
                return

            _logger.info("_monitor_response_handler - Received monitor response status: {}, message.transaction:{}, "
                         "len(message.object):{}".format(status, message.transaction, len(message.object)))

            delta_sync_msg_objs = []
            if not self._full_sync_transaction_done:
                # full sync transaction
                self._full_sync_msg_objs.extend(message.object)
                _logger.info("_monitor_response_handler - received full sync msg.")
                if not message.transaction.transaction_end:
                    _logger.info("_monitor_response_handler - full sync msg is not end.")
                    return
                _logger.info("_monitor_response_handler - full sync msg is end.")
                self._full_sync_transaction_done = True
                converted_fullsync_objs = MsgCache.update_with_full_msg(self._full_sync_msg_objs)
                self._full_sync_msg_objs.clear()
                delta_sync_msg_objs.extend(converted_fullsync_objs)
            else:
                # delta sync transaction
                for obj in message.object:
                    if MsgCache.update_with_delta_msg(obj):
                        delta_sync_msg_objs.append(obj)

            # perform delta sync work
            _logger.info("_monitor_response_handler - perform delta sync work.")
            for obj in delta_sync_msg_objs:
                if obj.object_type == types_pb2.PROFILE:
                    self._handle_profile_msg(obj)
                else:
                    self._handle_global_config_msg(obj)

        except Exception as e:
            trace = traceback.format_exc()
            _logger.error("_monitor_response_handler exception:{}, trace:{}".format(e, trace))

    def stop(self):
        """
        Close the connection to nestdb server
        """
        _logger.info("stopping nestdb client")
        self._keep_running = False

    @property
    def is_running(self):
        return self._keep_running

    def _destroy_rpc_stub(self):
        if self._stub:
            try:
                self._stub.Destroy()
                self._stub = None
                _logger.info("rpc stub destroyed")
            except Exception as e:
                trace = traceback.format_exc()
                _logger.error("Error in destroy rpc stub, Exception:{}, trace:{}".format(e, trace))
