# Copyright 2013-2021 VMware, Inc.  All rights reserved.
# -*- mode: Python; coding: utf-8 -*-
# **********************************************************
#
"""
Implementation of appliance SNMP configuration API.
"""

__author__ = "VMware, Inc."

import sys
import shutil
import logging
import os
import re
import json
import traceback
import subprocess as sb
from datetime import datetime
from xml.dom.minidom import parse, Node, Document, parseString
import time
import comm as cm
import cvm_fw
from cvm_fw import null_dev, read_log, DEFAULT_SNMP_PORT, DEF_TRAP_PORT
from pkg_defs import *

logger = logging.getLogger(__name__)

class ConfigError(Exception):
    def __init__(self, _code):
        self.code = _code

# linux file permissions -rw------
CFG_FILE_PERMS = 0o600

# SLES12 uses systemd, which performs both chkconfig and init.d functions.
SERVICE_CMD = ['/bin/systemctl']

# were to have output from SERVICE_CMD end up
errlog_path = "/var/log/vmw_snmp_service.log"

# SNMP control channel
SNMPD_CONTROL_PATH = "/var/run/snmp.ctl"
CONFIG_FILE = '/etc/vmware/snmp/snmp.xml'
TMP_CONFIG_FILE = "%s.tmp.%d" % (CONFIG_FILE, os.getpid())

# see perforce/bora vicfg-snmp for same regular expression
v1trp_re = re.compile(r'(?P<host>[\d\w][\w\d\.:-]*)(@(?P<port>\d+)){0,1}[\s/](?P<comm>[\x21-\x7e]+)')

# see: esxcli system snmp set -u, format: username/-|localized auth key/-|localized priv key/[auth|priv|none]
user_re = re.compile(r'(?P<user>[\W\w]+)[\s/](?P<auth>[-\w]+)[\s/](?P<priv>[-\w]+)[\s/](?P<lvl>\w+)')

# user lvl string must be one of these values:
sec_levels = {'auth': "", 'priv': "", 'none': ""}
# see: esxcli system snmp set -R  format: user/auth-proto/local-key/priv-proto/local-key/engine-id
remote_user_re = re.compile(r'(?P<user>\W\w+)[\s/](?P<authProto>[-\w]+)[\s/](?P<authKey>[-\w]+)[\s/](?P<privProto>[-\w]+)[\s/](?P<privKey>[-\w]+)[\s/](?P<engineid>\w+)')

# see: esxcli system snmp set -i format: ip@port/user/lvl/type
v3target_re = re.compile(r'(?P<ip>[\d\w][\w\d\.:-]+)(@(?P<port>\w+))?[\s/](?P<user>[\-\w]+)[\s/](?P<lvl>\w+)[\s/](?P<type>\w+)')

# keep this string mated to what rpm ships
def_cfg = '<config><snmpSettings><enable>false</enable><port>161</port><syscontact></syscontact><syslocation></syslocation><communities></communities><targets></targets><engineid></engineid><loglevel>warning</loglevel><authProtocol></authProtocol><privProtocol></privProtocol><v3users></v3users></snmpSettings></config>'

class Structure:
   pass

def _no_stats(reason):
     obj = Structure()
     obj.reason = reason
     obj.worstrtimelast = "n/a"
     obj.avgresponsetime = ""
     obj.worstresponsetime = ""
     obj.inpkts = 0
     outpkts = 0
     usmstatsnotintimewindows = 0
     usmstatsunknownusernames = 0
     usmstatsunknownengineids = 0
     usmstatswrongdigests = 0
     usmstatsdecryptionerrors = 0
     inbadversions = 0
     inbadcommunitynames = 0
     inbadcommunityuses = 0
     inasnparseerrs = 0
     intoobigs = 0
     innosuchnames = 0
     inbadvalues = 0
     inreadonlys = 0
     ingenerrs = 0
     intotalreqvars = 0
     intotalsetvars = 0
     ingetrequests = 0
     ingetnexts = 0
     insetrequests = 0
     ingetresponses = 0
     intraps = 0
     outtoobigs = 0
     outnosuchnames = 0
     outbadvalues = 0
     outgenerrs = 0
     outgetrequests = 0
     outgetnexts = 0
     outsetrequests = 0
     outgetresponses = 0
     outtraps = 0
     silentdrops = 0
     avgvarbinds = 0
     maxvarbinds = 0
     return obj

def _default_config():
    '''load agent configuration from xml file
    '''
    obj = Structure()
    obj.name = "SNMPConfigReadOnly"
    obj.enable=False
    obj.port=int(DEFAULT_SNMP_PORT)
    obj.loglevel='warning'
    obj.syscontact='not-set'
    obj.syslocation='not-set'
    obj.notraps=list()
    obj.communities=list()
    obj.targets=list()
    obj.engineid='not-set'
    obj.authentication="none"
    obj.privacy="none"
    obj.users=list()
    obj.remoteusers=list()
    obj.v3targets=list()
    obj.pid="n/a"
    return obj


def _get_kvp(kvp, key, default):
    if key in kvp:
       return kvp[key]
    return default

def _get_value(node):
     ''' Return text value given an xml node'''
     content = []
     for child in node.childNodes:
        if child.nodeType == Node.TEXT_NODE:
           content.append(child.nodeValue.strip())
     return "".join(content).strip()

def _signal_snmp(codes):
    '''if agent is configured, send it one or more codes specified
    esx-snmp-spec.txt section 6.1
    '''
    status = False

    # check if we should even try to connect
    try:
        kvp = _load_xml_file()
        if kvp:
            denable = str(_get_kvp(kvp, ENABLE, "false"))
            if denable.lower() != "true":
                logger.debug("No signal sent, agent is not configured to run enable=%s" % denable)
                return True
    except Exception as newerr:
        logging.error("_signal_snmp: unable to determine if agent was running from config, msg=%s" % newerr)
        return status

    try:
        # XXX TODO remove this? cvm_fw.update_firewall(kvp)
        status = cm.signal(codes)
    except Exception as err:
        logger.error("_signal_snmp: cm.signal(%s) failed %s" % (codes, err))
        return False
    return status

def _update_value(new_value, kvp, key):
    ''' return new_value if it is set else return kvp[key] if it is set
        return value in serialized format for writing to file
    '''
    # these keys indicate fields that can have their values cleared out if the value passed in is: 'reset'
    reset_dict = {FILTER : "", COMMUNITIES : "", TARGETS : "", V3TARGETS : "", USERS : "",
                  REMOTE_USERS : ""}

    if key in kvp:
        logger.debug("existing value is %s for key %s" % (kvp[key], key))
        if isinstance(new_value, int) and new_value == 0:
            logger.debug("using existing for int")
            return str(kvp[key])
        if not isinstance(new_value, int) and len(new_value) == 0:
            logger.debug("using existing value non-int")
            if key == TARGETS:
                if isinstance(kvp[TARGETS], list):
                    return SEP.join(kvp[TARGETS])
            elif key == COMMUNITIES:
                if isinstance(kvp[COMMUNITIES], list):
                    return SEP.join(kvp[COMMUNITIES])
            elif key == FILTER:
                if isinstance(kvp[FILTER], list):
                    return SEP.join(kvp[FILTER])
            elif key == USERS:
                if isinstance(kvp[USERS], list):
                    return SEP.join(kvp[USERS])
            elif key == REMOTE_USERS:
                if isinstance(kvp[REMOTE_USERS], list):
                    return SEP.join(kvp[REMOTE_USERS])
            elif key == V3TARGETS:
                if isinstance(kvp[V3TARGETS], list):
                    return SEP.join(kvp[V3TARGETS])
            return kvp[key]
    if isinstance(new_value, list):
        new_value = SEP.join(new_value)

    # reset operation
    if key in reset_dict:
        if new_value == 'reset':
            logger.debug("reset request received for key '%s'" % key)
            return ""

    logger.debug("_update_value: returning new_value (%s) for key: %s" % (new_value, key))
    return str(new_value)

def _load_xml(node, cfg):
      ''' Transfer data from xml/node to dict/cfg'''
      cfg[V2C] = False
      if node.nodeType == Node.ELEMENT_NODE and node.nodeName == HEADER1:
         for item in node.childNodes:
            if item.nodeType == Node.ELEMENT_NODE and item.nodeName == HEADER2:
               for elem in item.childNodes:
                  if elem.nodeType == Node.ELEMENT_NODE:
                     if elem.nodeName == COMMUNITIES:
                        rawvalue = _get_value(elem)
                        if rawvalue is not None:
                           cfg[COMMUNITIES] = rawvalue.split(SEP)
                     elif elem.nodeName == TARGETS:
                        rawvalue = _get_value(elem)
                        if rawvalue is not None:
                           cfg[TARGETS] = rawvalue.split(SEP)
                     elif elem.nodeName == ENABLE:
                        rawvalue = _get_value(elem)
                        cfg[ENABLE] = rawvalue.lower()
                     elif elem.nodeName == PORT:
                        rawvalue = _get_value(elem)
                        cfg[PORT] = int(rawvalue.strip())
                     elif elem.nodeName == FILTER:
                        rawvalue = _get_value(elem)
                        if rawvalue is not None:
                           cfg[FILTER] = rawvalue.split(SEP)
                     elif elem.nodeName == LOG_LEVEL:
                        rawvalue = _get_value(elem)
                        cfg[LOG_LEVEL] = rawvalue.strip()
                     elif elem.nodeName == SYSCONTACT:
                        rawvalue = _get_value(elem)
                        cfg[SYSCONTACT] = rawvalue.strip()
                     elif elem.nodeName == SYSLOCATION:
                        rawvalue = _get_value(elem)
                        cfg[SYSLOCATION] = rawvalue.strip()
                     elif elem.nodeName == ENGINE_ID: # SNMPv3 config
                        rawvalue = _get_value(elem)
                        cfg[ENGINE_ID] = rawvalue.strip()
                     elif elem.nodeName == AUTH_PROTO:
                        rawvalue = _get_value(elem)
                        cfg[AUTH_PROTO] = rawvalue.strip()
                        if cfg[AUTH_PROTO] == "":
                            cfg[AUTH_PROTO] = "usmNoAuthProtocol"
                     elif elem.nodeName == PRIV_PROTO:
                        rawvalue = _get_value(elem)
                        cfg[PRIV_PROTO] = rawvalue.strip()
                        if cfg[PRIV_PROTO] == "":
                            cfg[PRIV_PROTO] = "usmNoPrivProtocol"
                     elif elem.nodeName == USERS:
                        rawvalue = _get_value(elem)
                        if rawvalue is not None:
                            cfg[USERS] = rawvalue.split(SEP)
                     elif elem.nodeName == REMOTE_USERS:
                        rawvalue = _get_value(elem)
                        if rawvalue is not None:
                            cfg[REMOTE_USERS] = rawvalue.split(SEP)
                     elif elem.nodeName == V3TARGETS:
                        rawvalue = _get_value(elem)
                        if rawvalue is not None:
                            cfg[V3TARGETS] = rawvalue.split(SEP)
                     elif elem.nodeName == V2C:
                        cfg[V2C] = True

def _parse_tgt(target):
    '''convert string target to a pint structure'''
    if not len(target):
        return
    rslt = v1trp_re.match(target)
    if rslt is None:
        logger.error("_parse_tgt regular expression failed for input '%s', ignoring" % target)
        return None
    if rslt.group('port'):
        val = rslt.group('port')
    else:
        val = DEF_TRAP_PORT
    if not rslt.group('host'):
        logger.error("_parse_tgt missing mandatory host arg in '%s'" % target)
        return None
    if not rslt.group('comm'):
        logger.error("_parse_tgt missing mandatory community arg in '%s'" % target)
        return None
    tgt = pint.Structure('SNMPv1TrapTarget',
                          dict(ip= rslt.group('host'),
                               port=int(val),
                               community=rslt.group('comm')))
    return tgt

def _notification_type(cfg_val):
    '''/etc/vmware/snmp/snmp.xml stores type as numeric 1=trap,2=inform, convert to pint enum'''
    cfg_val = str(cfg_val)
    if cfg_val == '1':
        return "trap"
    elif cfg_val == '2':
        return "inform"
    logger.error("_notification_type not recognized '%s' expected 1 or 2" % cfg_val)
    raise Exception("Error interpreting the notification type %s" % cfg_val)

def _from_notification_type(pint_val):
    '''/etc/vmware/snmp/snmp.xml stores type as numeric 1=trap,2=inform, convert from pint enum'''
    pint_val = str(pint_val)
    if pint_val == 'trap':
        return "1"
    elif pint_val == 'inform':
        return "2"
    logger.error("_from_notification_type not recognized '%s' expected inform or trap" % pint_val)
    raise Exception("Error interpreting the notification type %s" % pint_val)

def _parse_v3target(item, out_pint):
    '''extract string and place values into pint/tgt
    item format: fc00:10:20:102::103@1162 joe noAuthNoPriv 2
    '''
    rslt = v3target_re.match(item)
    if rslt:
        try:
            out_pint.type = _notification_type(rslt.group('type'))
            out_pint.sec_level = _from_v3_level(rslt.group('lvl'))
            out_pint.ip = rslt.group('ip')
            if rslt.group('port'):
                out_pint.port = int(rslt.group('port'))  #PR 1254490
            out_pint.user = rslt.group('user')
            return True
        except Exception as err:
            logger.error("_parse_v3target: input '%s' failed: %s" % (item, err))
            return False
    else:
        logger.error("_parse_v3target: parse '%s' failed" % item)
    return False

def _v3tgt_to_pint(tgts):
    '''Convert array of strings and return array of Pint SNMPv3Target structs
    '''
    logger.debug("_v3tgt_to_pint: tgts=%s" % tgts)
    out = []

    if not isinstance(tgts, list):
        tgts = [tgts]

    try:
       if len(tgts[0]) == 0:
           return out
    except Exception:
        return out

    for item in tgts:
        logger.debug("_v3tgt_to_pint: converting item=%s" % item)
        tgt = pint.Structure("SNMPv3Target",
                              dict(type="trap",
                              ip="",
                              port=int(DEF_TRAP_PORT),
                              user="",
                              engineid=""))

        logger.debug("_v3tgt_to_pint: parsing item=%s" % item)
        if _parse_v3target(item, tgt):
            out.append(tgt)

    logger.debug("_v3tgt_to_pint: converted %d v3 targets out of %d" % (len(out), len(tgts)))
    return out

def _set_pid():
   ''' '''
   try:
       app_args = ["/bin/pidof", "snmpd"]
       report = sb.check_output(app_args, stdin=null_dev, stderr=null_dev).decode("utf-8")
       if report == "":
           return "running"
       return report.rstrip()
   except Exception:
       return "n/a"

def _convert(kvp):
    ''' generate a Pint structure from dict/KVP '''
    logging.info("_convert: kvp=%s" % kvp)
    cfg = _default_config()
    cfg.enable = (_get_kvp(kvp, ENABLE, str(cfg.enable).lower()) == "true")
    cfg.port = _get_kvp(kvp, PORT, cfg.port)
    cfg.communities = _get_kvp(kvp, COMMUNITIES, cfg.communities)
    trps = list()
    for tgt in kvp[TARGETS]:
        stgt = _parse_tgt(tgt)
        if stgt:
            trps.append(stgt)
    cfg.targets = trps
    cfg.syscontact = _get_kvp(kvp, SYSCONTACT, cfg.syscontact)
    cfg.syslocation = _get_kvp(kvp, SYSLOCATION, cfg.syslocation)
    cfg.loglevel = _get_kvp(kvp, LOG_LEVEL, cfg.loglevel)
    cfg.notraps = _get_kvp(kvp, FILTER, cfg.notraps)
    # SNMPv3 specific
    cfg.engineid = _get_kvp(kvp, ENGINE_ID, cfg.engineid)
    cfg.authentication = _convert_auth_proto_to_pint(_get_kvp(kvp, AUTH_PROTO, "usmNoAuthProtocol"))
    cfg.privacy = _priv_to_pint(_get_kvp(kvp, PRIV_PROTO, "usmNoPrivProtocol"))
    cfg.users = _user_to_pint(_get_kvp(kvp, USERS, cfg.users))
    cfg.remoteusers = _remote_user_to_pint(_get_kvp(kvp, REMOTE_USERS, cfg.remoteusers))
    cfg.v3targets = _v3tgt_to_pint(_get_kvp(kvp, V3TARGETS, cfg.v3targets))
    cfg.pid = _set_pid()
    logging.info("_convert: completed conversion %s" % str(cfg))
    return cfg

def _load_xml_file():
    '''return dict of key/values from xml file'''
    if os.path.getsize(CONFIG_FILE) > 0:
        doc = parse(CONFIG_FILE)
        kvp = dict()
        _load_xml(doc.firstChild, kvp)
        return kvp
    return None

def load_config():
    doc = None
    try:
        kvp = _load_xml_file()
        if kvp:
            return _convert(kvp)
        else: #
            logging.error('%s is empty, regenerating default values' % CONFIG_FILE)
            with open(CONFIG_FILE, "w") as fp:
                fp.write(def_cfg)
    except Exception as err:
        if doc:
            doc.unlink()
        msg = 'load_config: failed to parse %s, reason %s' % (CONFIG_FILE, str(err))
        logging.error(msg)
        tb = traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
        logging.error(msg)
        for line in tb:
            logging.error(line.rstrip())
        logging.error(tb)
        raise Exception(msg)

def _toggle_agent_status(truth_value):
    '''truth value is string true or false  '''
    desired_state = None
    enabled = None
    if truth_value == "true":
        desired_state = "true"
    elif truth_value == "false":
        desired_state = "false"
    else:
       raise Exception("Internal error invalid argument not true or false '%s'" % truth_value)

    try:
        kvp = _load_xml_file()
        if kvp:
            enabled = kvp.get(ENABLE, "false")
            logging.debug("current state is '%s' desired state is '%s'" % (enabled, desired_state))
            if enabled.lower() != desired_state:
               kvp[ENABLE] = truth_value
               _save_config(kvp)
            if cvm_fw.have_firewall():
                  cvm_fw.update_firewall(kvp)
            else:
               logger.debug("_toggle_agent_status: agent was '%s' nothing to change, value: %s" % (enabled, enabled))
        else:
            logger.error("_toggle_agent_status: _load_xml_file failed, no change made to agent config")
    except Exception as err:
        msg = '_toggle_agent_status: update enable state failed, reason %s' % str(err)
        logging.error(msg)
        tb = traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
        for line in tb:
            logging.error(line.rstrip())
        logging.error(tb)
        raise Exception(msg)

def _run_service(cmd):
    '''run service cmd to start/stop snmpd'''
    try:
       with open(errlog_path, 'w') as errlog:
           logger.debug("_run_service: invoking command: %s" % " ".join(cmd))
           sb.check_call(cmd, stdin=null_dev, stderr=errlog, stdout=errlog)
           logger.debug("_run_service: completed normally command: %s" % " ".join(cmd))
    except Exception as err:
        logmsg = read_log(errlog_path)
        msg = "_run_service: command '%s' failed {%s}, stdout={%s}" % (" ".join(cmd), err, logmsg)
        logger.error(msg)
        raise Exception(msg)

def enable():
    '''start agent running'''
    global SERVICE_CMD
    _toggle_agent_status('true')
    app_args = SERVICE_CMD + ["start", "snmpd"]
    _run_service(app_args)

def disable():
    ''' set agent off'''
    global SERVICE_CMD
    logging.debug("processing disable request.")
    _toggle_agent_status('false')
    app_args = SERVICE_CMD + ["stop", "snmpd"]
    _run_service(app_args)

def limits():
    ''' pass back structure with agent configuration limits
    see: /bora/lib/vmkctl/system/SnmpAgentConfigImpl.cpp#MAX_COMMUNITY_LEN
    '''
    lim = pint.Structure('SNMPLimits',
                          dict(maxCommunities=int(MAX_COMMUNITIES),
                               maxTrapDestinationsV1=int(MAX_TRAP_TARGETS),
                               maxDestinationsV3=int(MAX_TRAP_TARGETS),
                               maxNotificationFilters=int(MAX_FILTERS),
                               maxCommunityLength=int(MAX_COMMUNITY_LEN),
                               maxBufferSize=int(SR_DEFAULT_MAX_MSG_SIZE)))
    return lim

def _serialize_v1_targets(v1tgts):
    '''v1tgts is array of pint struct of type 'SNMPv1TrapTarget',
       return serialized string form as defined in esx-snmp-spec.txt section 6.1
    '''
    if v1tgts[0] == 'reset':
        return v1tgts
    if len(v1tgts) == 1 and v1tgts[0] == "":
        return ""
    cnt = 0
    spec = list()
    host = None
    comm = None
    for item in v1tgts:
        port = DEF_TRAP_PORT
        rslt = v1trp_re.match(item)
        if rslt is None:
            ex = Exception("Invalid target format, expected: ip-or-hostname[@port]/community[,...]")
            ex.bad_v1t = True
            raise ex
        if rslt.group('port'):
            port = rslt.group('port')
        else:
            port = port
        if not rslt.group('host'):
            logger.error("_serialize_v1_targets  missing mandatory host arg in '%s'" % item)
            continue
        else:
            host = rslt.group('host')
        if not rslt.group('comm'):
            logger.error("_serialize_v1_targets  missing mandatory community arg in '%s'" % item)
            continue
        else:
            comm = rslt.group('comm')
        sport = "@%s" % port
        spec.append("%s%s %s" % (host, sport, comm))
        cnt += 1
    logger.debug("_serialize_v1_targets: processed %d entries" % cnt)
    return SEP.join(spec)

def _parse_user(in_user, out_pint):
    '''extract four fields from in and place them in pint SNMPUser/out_pint
       in format: jane - - none
    '''

    if len(in_user) == 0:
        return False

    item = user_re.match(in_user)
    if item:
        try:
            out_pint.username = item.group('user')
            out_pint.sec_level = _from_v3_level(item.group('lvl'))
            out_pint.auth_key = item.group('auth')
            out_pint.priv_key = item.group('priv')
            return True
        except Exception as err:
            logger.error("_parse_user: parse '%s' failed: %s" % (in_user, err))
            return False
    else:
        logger.error("_parse_user: parse '%s' failed" % in_user)
    return False

def _user_to_pint(users):
    ''' convert user xml string to pint struct SNMPUser '''
    logger.debug("_user_to_pint: users=%s" % users)
    out = []
    if not isinstance(users, list):
        users = [users]

    if len(users) == 1 and users[0] == "":
        logger.debug("_users_to_pint: no users defined")
        return out

    for item in users:
        logger.debug("_user_to_pint: converting item=%s" % item)
        user = pint.Structure("SNMPUser",
                              dict(username="",
                                   sec_level="none",
                                   auth_key="",
                                   priv_key=""))
        logger.debug("_user_to_pint: parsing item=%s" % item)
        if _parse_user(item, user):
            out.append(user)

    logger.debug("_user_to_pint: converted %d users out of %d" % (len(out), len(users)))
    return out


def _convert_auth(authstr):
    ''' convert pint authentication enum to xml config file value, see esx-snmp-spec.xml sec 6.1'''
    if authstr:
        if authstr.upper() == 'SHA1':
            return 'usmHMACSHAAuthProtocol'
        elif authstr.upper() == 'MD5':
            return 'usmHMACMD5AuthProtocol'
        else:
           return 'usmNoAuthProtocol'

    return authstr

def _convert_auth_proto_to_pint(ap):
    ''' return pint enum for authentication field. They must be lowercase to match pint.py SNMPAuthProto keys'''
    if ap == 'usmNoAuthProtocol':
       return 'none'
    elif ap == "":
       return 'none'
    elif ap == 'usmHMACSHAAuthProtocol':
       return 'SHA1'
    elif ap == 'usmHMACMD5AuthProtocol':
       return 'MD5'
    logger.error("_convert_auth_proto_to_pint: unexpected auth proto value '%s'" % ap)
    raise Exception("Invalid authentication protocol found in config file, use reset command to clear.")

def _convert_auth_proto(protocol):
    ''' change pint enum to file format as per esx-snmp-spec.txt '''
    if protocol.upper() == 'NONE':
        return 'usmNoAuthProtocol'
    elif protocol.upper() == 'SHA1':
        return 'usmHMACSHAAuthProtocol'
    elif protocol.upper() == 'MD5':
        return 'usmHMACMD5AuthProtocol'
    raise ConfigError(16)


def _convert_priv_proto_to_pint(priv):
    '''return pint enum for privacy field, must be lowercase to match pint.py enum SNMPPrivProto keys'''
    if priv == 'usmNoPrivProtocol':
       return 'none'
    elif priv == '':
       return 'none'
    elif priv == 'usmAESCfb128PrivProtocol':
       return 'AES128'
    logger.error("_convert_priv_proto_to_pint: unexpected auth proto value '%s'" % priv)
    raise Exception("Invalid authentication protocol found in config file, use reset command to clear.")

def _priv_to_pint(priv_str):
    ''' convert priv config to pint enum '''
    if priv_str:
       if priv_str == 'usmAESCfb128PrivProtocol':
           return 'AES128'
    return 'none'

def _convert_priv(privstr):
    ''' convert pint privacy enum to xml config file value see esx-snmp-spec.xml sec 6.1'''
    if privstr:
        if privstr.upper() == 'NONE':
            return 'usmNoPrivProtocol'
        if privstr.upper() == 'AES128':
            return 'usmAESCfb128PrivProtocol'
        else:
           return ''
    return privstr


def _parse_remote_user(in_user, out_pint):
    '''extract six fields from in and place them in pint SNMPRemoteUser/out_pint
       in format: jane - - none
    '''
    item = remote_user_re.match(in_user)
    if item:
        try:
            out_pint.username = item.group('user')
            out_pint.authentication = _convert_auth_proto_to_pint(item.group('authProto'))
            out_pint.auth_key = item.group('authKey')
            out_pint.privacy = _convert_priv_proto_to_pint(item.group('privProto'))
            out_pint.priv_key = item.group('privKey')
            out_pint.engineid = item.group('engineid')
            return True
        except Exception as err:
            logger.error("_parse_user: parse '%s' failed: %s" % (in_user, err))
            return False
    else:
        logger.error("_parse_user: parse '%s' failed" % in_user)
    return False


def _remote_user_to_pint(users):
    ''' convert remote user xml string to pint struct SNMPUser '''
    logger.debug("_user_to_pint: users=%s" % users)
    out = []
    if not isinstance(users, list):
        users = [users]
    for item in users:
        logger.debug("_remote_user_to_pint: converting item=%s" % item)
        user = pint.Structure("SNMPRemoteUser",
                              dict(username="",
                                   sec_level="none",
                                   authentication="none",
                                   auth_key="",
                                   privacy="none",
                                   priv_key="",
                                   engineid=""))
        logger.debug("_remote_user_to_pint: parsing item=%s" % item)
        if _parse_remote_user(item, user):
            out.append(user)

    logger.debug("_remote_user_to_pint: converted %d remote users out of %d" % (len(out), len(users)))
    return out


def _serialize_users(pint_users):
    '''convert Pint SNMPUsers array to array of strings for xml config file'''
    if pint_users[0] == 'reset':
        return pint_users
    if len(pint_users) == 1 and pint_users[0] == "":
        return ""
    out = []
    for item in pint_users:
        rslt = user_re.match(item)
        if rslt:
            # XXX TODO check length of name, length of auth/priv strings
            try:
                lvl = _convert_v3_level(rslt.group('lvl'))
                entry = "%s %s %s %s" % (rslt.group('user'), rslt.group('auth'), rslt.group('priv'), lvl)
                out.append(entry)
            except Exception as err:
                logger.error("_serialize_users: ignoring entry '%s' error %s" % (item, err))
        else:
            raise Exception("Invalid user account format, expected: userid/-|authHash/-|privHash/model[,...]")

    logger.debug("_serialize_users: converted %d out of %d users" % (len(out), len(pint_users)))
    return out

def _convert_priv_proto(protocol):
    ''' change pint enum to file format as per esx-snmp-spec.txt '''
    if protocol == 'none':
        return 'usmNoPrivProtocol'
    elif protocol.upper() == 'AES128':
        return 'usmAESCfb128PrivProtocol'
    raise ConfigError(16)

def _validate_log_level(level):
    '''throw if loglevel is not a string with proscribed value in snmp/pint.py'''
    logger.info("_validate_log_level: log level requested is: '%s'" % level)
    if level == 'debug' or level == 'info' or level == 'warning' or level == 'error':
       return
    logger.error("_validate_log_level: invalid log level '%s'" % level)
    raise ConfigError(16)

def _serialize_remote_users(pint_users):
    '''convert Pint SNMPUsers array to array of strings for xml config file'''
    if pint_users[0] == 'reset':
        return pint_users
    out = []
    for item in pint_users:
        rslt = remote_user_re.match(item)
        if rslt:
            # XXX TODO validate data received
            user = rslt.group('user')
            auth = _convert_auth_proto(rslt.group('authProto'))
            authKey = str(rslt.group('authKey'))
            priv = _convert_priv_proto(rslt.group('privProto'))
            privKey = str(rslt.group('privKey'))
            engineid = rslt.group('engineid')

            entry = "%s %s %s %s %s %s" % (user, auth, authKey, priv, privKey, engineid)
            out.append(entry)
        else:
            raise Exception( "Invalid remote-user account format, expected: user/auth-proto/-|authKey/priv-proto/-|privKey/engine-id[,...]")

    logger.debug("_serialize_remote_users: converted %d out of %d remote users" % (len(out), len(pint_users)))
    return out

def _from_v3_level(val):
    '''convert pint secuerity level to xml file format '''
    if val == 'noAuthNoPriv':
        return 'none'
    elif val == 'authNoPriv':
        return 'auth'
    elif val == 'authPriv':
        return 'priv'
    logger.error("_from_v3_level: unexpected security pint level '%s'" % val)
    raise Exception("Security level for notification not known %s" % val)

def _convert_v3_level(val):
    '''convert pint secuerity level to xml file format '''
    if val == 'none':
        return 'noAuthNoPriv'
    elif val == 'auth':
        return 'authNoPriv'
    elif val == 'priv':
        return 'authPriv'
    logger.error("_convert_v3_level: unexpected security pint level '%s'" % val)
    raise Exception("Security level for notification not known %s" % val)

def _serialize_v3targets(pint_tgts):
    '''convert SNMPConfig.v3targets array to format used in xml config file'''
    if pint_tgts[0] == 'reset':
        return pint_tgts
    out = []
    for item in pint_tgts:
        rslt = v3target_re.match(item)
        if rslt:
            port = DEF_TRAP_PORT
            dtype = _from_notification_type(rslt.group('type'))
            sec_level = _convert_v3_level(rslt.group('lvl'))
            user = rslt.group('user')
            ip = rslt.group('ip')
            if rslt.group('port'):
               port = rslt.group('port')
            user = rslt.group('user')
            entry = "%s@%s %s %s %s" % (ip, port, user, sec_level, dtype)
            out.append(entry)
        else:
            ex = Exception("Invalid 3target format, expected: ip-or-hostname[@port]/remote-user/security-level/trap|inform[,...]")
            ex.bad_v3t = True
            raise ex

    logger.debug("_serialize_v3targets: converted %d out of %d v3 targets" % (len(out), len(pint_tgts)))
    return out

def _merge_config(kvp, pint_cfg):
    '''update dict/kvp (contents of /etc/vmware/snmp/snmp.xml) with values from SNMPConfig/pint_cfg
       also peform validation of input from user here.
    '''
    logger.debug("_merge_config: merging change request")
    kvp[AUTH_PROTO] = _update_value(_convert_auth(pint_cfg.authentication), kvp, AUTH_PROTO)
    kvp[COMMUNITIES] = _update_value(pint_cfg.communities, kvp, COMMUNITIES)
    kvp[ENGINE_ID] = _update_value(pint_cfg.engineid, kvp, ENGINE_ID)
    kvp[LOG_LEVEL] = _update_value(pint_cfg.loglevel, kvp, LOG_LEVEL)
    kvp[FILTER] = _update_value(pint_cfg.notraps, kvp, FILTER)
    if pint_cfg.port != -1:
        kvp[PORT] = _update_value(pint_cfg.port, kvp, str(PORT))
    kvp[PRIV_PROTO] = _update_value(_convert_priv(pint_cfg.privacy), kvp, PRIV_PROTO)
    kvp[SYSCONTACT] = _update_value(pint_cfg.syscontact, kvp, SYSCONTACT)
    kvp[SYSLOCATION] = _update_value(pint_cfg.syslocation, kvp, SYSLOCATION)
    if pint_cfg.v2c != None:
        kvp[V2C] = _update_value(pint_cfg.v2c, kvp, V2C)

    if len(pint_cfg.targets) > 0:
        tval = _serialize_v1_targets(pint_cfg.targets)
    else:
        tval = ""
    kvp[TARGETS] = _update_value(tval, kvp, TARGETS)

    if len(pint_cfg.users) > 0:
        tval = _serialize_users(pint_cfg.users)
    else:
        tval = ""
    kvp[USERS] = _update_value(tval, kvp, USERS)

    if len(pint_cfg.remoteusers) > 0:
        tval = _serialize_remote_users(pint_cfg.remoteusers)
    else:
        tval = ""
    kvp[REMOTE_USERS] = _update_value(tval, kvp, REMOTE_USERS)

    if len(pint_cfg.v3targets) > 0:
        tval = _serialize_v3targets(pint_cfg.v3targets)
    else:
        tval = ""
    kvp[V3TARGETS] = _update_value(tval, kvp, V3TARGETS)

def _save_config(kvp):
    '''write contents of kvp dict to a temp file, then rename it to CONFIG_FILE'''
    if not kvp:
        raise Exception("_save_config: invalid argument passed")
    logger.debug("_save_config: generating replacement config {%s}" % kvp)
    doc = Document()
    root = doc.createElement("config")
    doc.appendChild(root)
    head = doc.createElement("snmpSettings")
    root.appendChild(head)

    ena = doc.createElement(ENABLE)
    head.appendChild(ena)
    ena.appendChild(doc.createTextNode(str(kvp.get(ENABLE, "false")).lower()))

    port = doc.createElement(PORT)
    head.appendChild(port)
    port.appendChild(doc.createTextNode(str(kvp.get(PORT, DEFAULT_SNMP_PORT))))

    loglevel = doc.createElement(LOG_LEVEL)
    head.appendChild(loglevel)
    loglevel.appendChild(doc.createTextNode(kvp.get(LOG_LEVEL, "warning")))

    # Append v2c as flags if set true
    vopt = doc.createElement(V2C)
    if str(kvp.get(V2C, False)) == "True":
       head.appendChild(vopt)

    contact = doc.createElement(SYSCONTACT)
    head.appendChild(contact)
    contact.appendChild(doc.createTextNode(kvp.get(SYSCONTACT, "not-set")))

    location = doc.createElement(SYSLOCATION)
    head.appendChild(location)
    location.appendChild(doc.createTextNode(kvp.get(SYSLOCATION, "not-set")))

    eid = doc.createElement(ENGINE_ID)
    head.appendChild(eid)
    xid = kvp.get(ENGINE_ID, "")
    if xid:
        eid.appendChild(doc.createTextNode(xid))

    ap = doc.createElement(AUTH_PROTO)
    head.appendChild(ap)
    proto = kvp.get(AUTH_PROTO, "")
    if proto:
        ap.appendChild(doc.createTextNode(proto))

    pp = doc.createElement(PRIV_PROTO)
    head.appendChild(pp)
    proto = kvp.get(PRIV_PROTO, "")
    if proto:
        pp.appendChild(doc.createTextNode(proto))

    logger.debug("setting communities [%s]" % kvp.get(COMMUNITIES, ""))
    com_e = doc.createElement(COMMUNITIES)
    head.appendChild(com_e)
    comm = kvp.get(COMMUNITIES, "")
    if len(comm) == 1 and comm[0] == "":
       pass
    else:
       if isinstance(comm, list):
           comm = SEP.join(comm)
       com_e.appendChild(doc.createTextNode(comm))
    del comm

    logger.debug("setting trap targets [%s]" % kvp.get(TARGETS, ""))
    tgt = doc.createElement(TARGETS)
    head.appendChild(tgt)
    tt = kvp.get(TARGETS, "")
    if len(tt) == 1 and tt[0] == "":
       pass
    else:
        if isinstance(tt, list):
             tt = SEP.join(tt)
        tgt.appendChild(doc.createTextNode(tt))
    del tgt
    del tt

    logger.debug("setting notification filters [%s]" % kvp.get(FILTER, ""))
    dfilter = doc.createElement(FILTER)
    head.appendChild(dfilter)
    flt = kvp.get(FILTER, "")
    if len(flt) == 1 and flt[0] == "":
       pass
    else:
        if isinstance(flt, list):
           flt = SEP.join(flt)
        dfilter.appendChild(doc.createTextNode(flt))
    del dfilter
    del flt

    logger.debug("setting users [%s]" % kvp.get(USERS, ""))
    users = doc.createElement(USERS)
    head.appendChild(users)
    v3u = kvp.get(USERS, "")
    if len(v3u) == 1 and v3u[0] == "":
       pass
    else:
        if isinstance(v3u, list):
           v3u = SEP.join(v3u)
        users.appendChild(doc.createTextNode(v3u))
    del users
    del v3u

    logger.debug("setting remote users [%s]" % kvp.get(REMOTE_USERS, ""))
    rusers = doc.createElement(REMOTE_USERS)
    head.appendChild(rusers)
    v3r = kvp.get(REMOTE_USERS, "")
    if len(v3r) == 1 and v3r[0] == "":
       pass
    else:
        if isinstance(v3r, list):
           v3r = SEP.join(v3r)
        rusers.appendChild(doc.createTextNode(v3r))
    del rusers
    del v3r

    logger.debug("setting v3targets [%s]" % kvp.get(V3TARGETS, ""))
    v3tgt = doc.createElement(V3TARGETS)
    head.appendChild(v3tgt)
    tgts = kvp.get(V3TARGETS, "")
    if len(tgts) == 1 and tgts[0] == "":
       pass
    else:
        if isinstance(tgts, list):
           tgts = SEP.join(tgts)
        v3tgt.appendChild(doc.createTextNode(tgts))
    del v3tgt
    del tgts

    try:
        with open(TMP_CONFIG_FILE, "w") as cfgFile:
            doc.writexml(cfgFile)
        doc.unlink()
        msg = "_save_config: renaming config file %s to %s" % (TMP_CONFIG_FILE, CONFIG_FILE)
        logger.debug(msg)
        os.chmod(TMP_CONFIG_FILE, CFG_FILE_PERMS)
        shutil.move(TMP_CONFIG_FILE, CONFIG_FILE)
        logger.info("_save_config: %s updated." % CONFIG_FILE)
    except IOError as err:
        if doc is not None:
            doc.unlink()
        msg = '_save_config: move %s to %s failed %s' % \
                 (TMP_CONFIG_FILE, CONFIG_FILE, str(err))
        logging.error(msg)
        raise Exception(msg)


def set_config(pint_cfg):
    '''write to persistent store and inform snmp process if running'''
    func = "n/a"
    try:
        func = "_load_xml_file()"
        kvp = _load_xml_file()
        func = "_merge_config()"
        _merge_config(kvp, pint_cfg)
        func = "_save_config()"
        _save_config(kvp)
        logger.debug("set_config: config file updated")
    except Exception as err:
        msg = "%s: %s" % (func, str(err))
        if os.path.exists('/tmp/trace-vicfg-snmp'):
           tb = traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
           for line in tb:
               logging.error(line.rstrip())
           logger.error(tb)
        ex = Exception(msg)
        if hasattr(err, 'bad_v3t'):
           ex.bad_v3t = True
        if hasattr(err, 'bad_v1t'):
           ex.bad_v1t = True
        raise ex

    try:
        # see /bora/apps/snmp/doc/esx-snmp-spec.txt sec 3.2
        _signal_snmp('U')
    except Exception as err:
        msg = "set_config: cm.signal('U') failed: %s" % str(err)
        logger.error(msg)
        raise Exception(msg)

    logger.info("SNMP agent configuration transaction completed normally.")


def reset_config():
    ''' reset configuration to default settings'''
    disable()
    try:
        with open(TMP_CONFIG_FILE, "w") as fp:
            fp.write(def_cfg)
            fp.close()
    except Exception as err:
       msg = "reset_config: Write default configuration file %s failed: %s" % (CONFIG_FILE, str(err))
       logger.error(msg)
       raise Exception(msg)

    try:
         os.chmod(TMP_CONFIG_FILE, CFG_FILE_PERMS)
         shutil.move(TMP_CONFIG_FILE, CONFIG_FILE)
    except Exception as err:
        msg = "reset_config: rename(%s->%s) failed: %s" % (TMP_CONFIG_FILE, CONFIG_FILE, str(err))
        logger.error(msg)
        raise Exception(msg)
    logger.info("SNMP Agent config has been reset to factory settings.")

def _count_targets(cfgstr):
    '''Count number of destination targets '''
    if not isinstance(cfgstr, list):
       return 0
    if len(cfgstr) == 1 and cfgstr[0] == "":
       return 0
    return len(cfgstr)

def test_config(raw = False):
    '''send a test notification (trapv1, trapv3, inform) or report why it can not be sent'''
    try:
        if not raw:
           results = pint.Structure('SNMPTestResults',
                                 dict(success=False, message="n/a"))
        else:
           results = dict()
        kvp = _load_xml_file()
        tgt_cnt = _count_targets(kvp.get(TARGETS, []))
        tgt_cnt += _count_targets(kvp.get(V3TARGETS, []))
        logger.debug("test_config: _count_targets(v1,v3) returned %s" % tgt_cnt)
        if tgt_cnt == 0:
            if not raw:
               results.success = False
               results.message = "Must first configure at least one v1|v2c|v3 trap target."
            else:
               results['success'] = False
               results['message'] = "Must first configure at least one v1|v2c|v3 trap target."
            return results

        if kvp.get(ENABLE, "false") == "true":
           logger.debug("test_config: agent is enabled")
           logger.debug("test_config: agent has %d targets configured" % tgt_cnt)
           logger.debug("Requesting agent to send test notification....")
           if _signal_snmp('T'): # see /bora/apps/snmp/doc/esx-snmp-spec.txt sec 3.2
               if not raw:
                   results.success = True
               else:
                   results['success'] = True

               plural = "s"
               if tgt_cnt == 1:
                    plural = ""
               if not raw:
                  results.message = "A warmStart notification was sent to %d target%s." % (tgt_cnt, plural)
               else:
                  results['message'] = "A warmStart notification was sent to %d target%s." % (tgt_cnt, plural)
               logger.debug("Request to generate diagnostic notification completed normally.")
           else:
               if not raw:
                  results.success = False
                  results.message = "Internal error - command to snmpd agent to send notification failed. See logs."
               else:
                  results['success'] = False
                  results['message'] = "Internal error - command to snmpd agent to send notification failed. See logs."
           return results
        else:
            if not raw:
               results.success = False
               results.message = "Must first enable agent."
            else:
               results['success'] = False
               results['message'] = "Must first enable agent."
            return results
    except Exception as err:
        msg = "test_config: internal error, failed: %s" % str(err)
        logger.error(msg)
        tb = traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
        for line in tb:
            logging.error(line.rstrip())
        logger.error(tb)
        raise Exception(msg)

def load_stats():
    '''retrieve snmp live performance metrics '''
    stats_path = "/tmp/snmp-stats"
    with open(stats_path) as fp:
        stats = json.load(fp);
        the_date = "n/a"
        at = stats.get('worstrtimelast', 0)
        if at != 0:
            timestamp = datetime.fromtimestamp(at)
            the_date = timestamp.strftime("%FT%TZ")
        obj = Structure()
        obj.sysuptime = "%s seconds" % (stats.get('sysUpTime', 0) / 100)
        obj.worstrtimelast = the_date
        obj.avgresponsetime = "%d ms" % stats.get('avgresponsetime',0)
        obj.worstresponsetime = "%d ms" % stats.get('worstresponsetime',0)
        obj.inpkts = stats.get('inpkts',5)
        obj.outpkts = stats.get('outpkts',6)
        obj.usmstatsnotintimewindows = stats.get('usmstatsnotintimewindows',11)
        obj.usmstatsunknownusernames = stats.get('usmstatsunknownusernames',12)
        obj.usmstatsunknownengineids = stats.get('usmstatsunknownengineids', 13)
        obj.usmstatswrongdigests = stats.get('usmstatswrongdigests',14)
        obj.usmstatsdecryptionerrors = stats.get('usmstatsdecryptionerrors',15)
        obj.inbadversions = stats.get('inbadversions',18)
        obj.inbadcommunitynames = stats.get('inbadcommunitynames',19)
        obj.inbadcommunityuses = stats.get('inbadcommunityuses',0)
        obj.inasnparseerrs = stats.get('inasnparseerrs',0)
        obj.intoobigs = stats.get('intoobigs',0)
        obj.innosuchnames = stats.get('innosuchnames',0)
        obj.inbadvalues = stats.get('inbadvalues',0)
        obj.ingenerrs = stats.get('ingenerrs',0)
        obj.outtoobigs = stats.get('outtoobigs',0)
        obj.outnosuchnames = stats.get('outnosuchnames',0)
        obj.outbadvalues = stats.get('outbadvalues',0)
        obj.outgenerrs = stats.get('outgenerrs',0)
        obj.outtraps = stats.get('outtraps',0)
        obj.silentdrops = stats.get('silentdrops',0)
        obj.avgvarbinds = stats.get('avgvarbinds',0)
        obj.maxvarbinds = stats.get('maxvarbinds',0)
        return obj

def get_stats(raw = False):
    '''get statistics report from snmpd or return the disabled report'''
    results = _no_stats("No stats available")
    try:
        kvp = _load_xml_file()
        if kvp.get(ENABLE, "false") == "false":
             if raw is False:
                return _no_stats("agent is disabled")
        # if agent running, ask it to generate report
        running = _set_pid()
        if running == "n/a":
           print("INFO: report statistics requires agent to be running. Check agent configuration.")
           return _no_stats("agent is not running")
        if not _signal_snmp('S'):
             if raw is False:
                return _no_stats("agent did not respond to request for statistics")
        tries = 3
        results = _no_stats("agent did not generate stats after request")
        while tries > 0:
            try:
                results = load_stats()
                break
            except:
                tries -= 1
                time.sleep(2)
        return results
    except Exception as err:
        msg = "get_stats: internal error, failed: %s" % str(err)
        logger.error(msg)
        tb = traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
        for line in tb:
            logging.error(line.rstrip())
        logger.error(tb)
    return _no_stats("General exception obtaining statistics")

def check_path(path):
    ''' verify path is valid '''
    if not os.path.exists(path):
       raise ConfigError(15)

def make_keys(pint_config):
    '''Given auth, priv secret run snmpd to generate localized keys
       See esx-snmp-spec.txt section 4.2.34  where it explains how to generate localized hashes
    default config will report: loaded from config, authProto=usmNoAuthProtocol privProto=usmNoPrivProtocol, eid=
    '''
    MIN_SECRET_LEN = 8
    try:
        doPriv = False
        kvp = _load_xml_file()
        ap = kvp.get(AUTH_PROTO, "usmNoAuthProtocol")
        pp = kvp.get(PRIV_PROTO, "usmNoPrivProtocol")
        eid = kvp.get(ENGINE_ID, "")
        logger.debug("make_keys: loaded from config, authProto=%s privProto=%s, eid=%s" % (ap, pp, eid))

        override = pint_config.raw_secret
        app_hash = pint_config.auth_hash
        priv_hash = pint_config.priv_hash

        logger.debug("ap=%s, pp=%s" % (ap, pp))

        if len(eid) == 0:
            logger.warning("engine ID not set yet, agent will generate one")

        if len(ap) == 0 or ap == "usmNoAuthProtocol":
            logger.debug("make_keys: authentication protocol must be configured first.")
            raise ConfigError(2)


        if override and len(app_hash) < MIN_SECRET_LEN:
            raise ConfigError(1)

        ap = _convert_auth_proto_to_pint(ap)

        app_args = ["/bin/snmpd", "-a", ap, "-s", app_hash]
        if override:
           app_args.append("-r")
        else:
           check_path(app_hash)

        if len(priv_hash) > 0:
            if len(pp) == 0 or pp == "usmNoPrivProtocol":
                logger.debug("make_keys: authentication protocol must be configured first.")
                raise ConfigError(4)
            if not override:
                check_path(priv_hash)
            if override and len(priv_hash) < MIN_SECRET_LEN:
                raise ConfigError(3)
            doPriv = True
            app_args.append("-p")
            pp = _priv_to_pint(pp).upper()
            app_args.append(pp)
            app_args.append("-s")
            app_args.append(priv_hash)
        logger.debug("running cmd: %s" % " ".join(app_args))
        try:
           with open(errlog_path, 'w') as errlog:
               report = sb.check_output(app_args, stdin=null_dev, stderr=errlog)
        except sb.CalledProcessError as err:
            logger.error("make_keys: generate hash failed: %s" % str(err))
            logger.error("make_keys: cmd stderr log was: %s" % read_log(errlog_path))
            raise Exception(str(err))

        auth_hash = ""
        priv_hash = ""
        # report should look like:
        #  <snmpv3><auth>478a217bd89562a25942a5f7daf8519</auth><priv>b3599b55e664a2adeef9c9c9e4c88da</priv></snmpv3>
        doc = parseString(report)
        items = doc.getElementsByTagName('auth')
        if items:
            auth_hash = _get_value(items[0])
        else:
            raise ConfigError(4)
        if doPriv:
            items = doc.getElementsByTagName('priv')
            priv_hash = _get_value(items[0])
        results = pint.Structure('SNMPHashResults',
                                 dict(auth_key=auth_hash,
                                      priv_key=priv_hash))

        return results

    except ConfigError:
        raise
    except Exception as err:
        msg = "make_keys: internal error, failed: %s" % str(err)
        logger.error(msg)
        tb = traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
        for line in tb:
            logging.error(line.rstrip())
        logger.error(tb)
        raise Exception(msg)
# end of file
